from __future__ import annotations

from pathlib import Path
from typing import Any

import pytest
import requests

from acp.discovery import DiscoveryClient
from acp.identity import AgentIdentity


class DummyResponse:
    def __init__(self, status_code: int, body: dict[str, Any] | None = None) -> None:
        self.status_code = status_code
        self._body = body

    def json(self) -> dict[str, Any]:
        if self._body is None:
            raise ValueError("No JSON body")
        return self._body


def test_discovery_resolves_identity_document_from_well_known(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    target = AgentIdentity.create("agent:shipping.bot@company.local")
    identity_document = target.build_identity_document(
        direct_endpoint="https://company.local/acp/inbox",
        relay_hints=["https://relay.company.local"],
        trust_profile="domain_verified",
        capabilities={"agent_id": target.agent_id},
    )
    well_known = {
        "agent_id": target.agent_id,
        "identity_document": "https://company.local/api/v1/acp/identity",
        "transports": {"http": {"endpoint": "https://company.local/acp/inbox"}},
        "version": "1.0",
        "security_profile": "https",
    }

    def fake_get(
        url: str,
        params: dict[str, str] | None = None,
        timeout: int = 5,
        **_kwargs: object,
    ) -> DummyResponse:
        if params is not None:
            return DummyResponse(404)
        if url == "https://company.local/.well-known/acp":
            return DummyResponse(200, well_known)
        if url == "https://company.local/api/v1/acp/identity":
            return DummyResponse(200, {"identity_document": identity_document})
        return DummyResponse(404)

    monkeypatch.setattr(requests, "get", fake_get)

    cache_path = tmp_path / "discovery_cache.json"
    client = DiscoveryClient(cache_path=cache_path)
    resolved = client.resolve(target.agent_id)
    assert resolved["agent_id"] == target.agent_id
    assert client.cache[target.agent_id].identity_document["agent_id"] == target.agent_id


def test_resolve_well_known_supports_base_url(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    target = AgentIdentity.create("agent:trading.bot@company.local")
    identity_document = target.build_identity_document(
        direct_endpoint="https://company.local/acp/inbox",
        relay_hints=[],
        trust_profile="domain_verified",
        capabilities={"agent_id": target.agent_id},
    )
    well_known = {
        "agent_id": target.agent_id,
        "identity_document": "/api/v1/acp/identity",
        "transports": {"http": {"endpoint": "https://company.local/acp/inbox"}},
        "version": "1.0",
        "security_profile": "https",
    }

    def fake_get(
        url: str,
        params: dict[str, str] | None = None,
        timeout: int = 5,
        **_kwargs: object,
    ) -> DummyResponse:
        if params is not None:
            return DummyResponse(404)
        if url == "https://company.local/.well-known/acp":
            return DummyResponse(200, well_known)
        if url == "https://company.local/api/v1/acp/identity":
            return DummyResponse(200, {"identity_document": identity_document})
        return DummyResponse(404)

    monkeypatch.setattr(requests, "get", fake_get)

    client = DiscoveryClient(cache_path=tmp_path / "cache.json")
    resolved = client.resolve_well_known("https://company.local", expected_agent_id=target.agent_id)
    assert resolved["well_known_url"] == "https://company.local/.well-known/acp"
    assert resolved["well_known"]["agent_id"] == target.agent_id
    assert resolved["identity_document"]["agent_id"] == target.agent_id



from __future__ import annotations

import json
from pathlib import Path
from typing import Any

import pytest

from acp.identity import AgentIdentity, write_identity
from acp.key_provider import VaultKeyProvider
from acp.messages import DeliveryOutcome, DeliveryState, SendResult
from acp_cli.main import main


def _seed_identity(storage_dir: Path, agent_id: str) -> None:
    identity = AgentIdentity.create(agent_id)
    identity_document = identity.build_identity_document(
        direct_endpoint="https://localhost:8443/api/v1/acp/messages",
        relay_hints=["https://relay.example"],
        trust_profile="self_asserted",
        capabilities={"agent_id": agent_id, "transports": ["direct", "relay"]},
    )
    write_identity(storage_dir, identity, identity_document)


def test_identity_show_reports_local_key_provider(tmp_path: Path, capsys) -> None:
    agent_id = "agent:key.local.show@demo"
    _seed_identity(tmp_path, agent_id)

    code = main(["--storage-dir", str(tmp_path), "--json", "identity", "show", "--agent-id", agent_id])
    assert code == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["ok"] is True
    assert payload["key_provider"]["provider"] == "local"


def test_identity_show_reports_vault_provider_metadata(tmp_path: Path, capsys) -> None:
    agent_id = "agent:key.vault.show@demo"
    _seed_identity(tmp_path, agent_id)

    code = main(
        [
            "--storage-dir",
            str(tmp_path),
            "--key-provider",
            "vault",
            "--vault-url",
            "https://vault.example",
            "--vault-path",
            "secret/data/acp/identities",
            "--json",
            "identity",
            "show",
            "--agent-id",
            agent_id,
        ],
    )
    assert code == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["ok"] is True
    assert payload["key_provider"]["provider"] == "vault"
    assert payload["key_provider"]["vault_url"] == "https://vault.example"
    assert payload["key_provider"]["vault_path"] == "secret/data/acp/identities"


def test_message_send_passes_vault_key_provider(monkeypatch: pytest.MonkeyPatch, tmp_path: Path, capsys) -> None:
    captured: dict[str, Any] = {}

    class FakeAgent:
        def send(self, **kwargs: Any) -> SendResult:
            captured["send_kwargs"] = kwargs
            recipients = kwargs["recipients"]
            return SendResult(
                operation_id="op-vault",
                message_id="msg-vault",
                message_ids=["msg-vault"],
                outcomes=[DeliveryOutcome(recipient=item, state=DeliveryState.DELIVERED) for item in recipients],
            )

    def fake_load_or_create(agent_id: str, **kwargs: Any) -> FakeAgent:
        captured["agent_id"] = agent_id
        captured["agent_kwargs"] = kwargs
        return FakeAgent()

    monkeypatch.setattr("acp_cli.message_commands.Agent.load_or_create", fake_load_or_create)

    code = main(
        [
            "--storage-dir",
            str(tmp_path),
            "--key-provider",
            "vault",
            "--vault-url",
            "https://vault.example",
            "--vault-path",
            "secret/data/acp/identities",
            "--json",
            "message",
            "send",
            "--from",
            "agent:sender@demo",
            "--to",
            "agent:receiver@demo",
            "--payload-json",
            '{"kind":"ping"}',
        ],
    )
    assert code == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["ok"] is True
    provider = captured["agent_kwargs"]["key_provider"]
    assert isinstance(provider, VaultKeyProvider)
    assert provider.vault_url == "https://vault.example"
    assert provider.vault_path == "secret/data/acp/identities"


def test_agent_status_reports_key_provider(tmp_path: Path, capsys) -> None:
    agent_id = "agent:key.status@demo"
    _seed_identity(tmp_path, agent_id)

    code = main(["--storage-dir", str(tmp_path), "--json", "agent", "status", "--agent-id", agent_id])
    assert code == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["ok"] is True
    assert payload["key_provider"]["provider"] == "local"


def test_vault_key_provider_missing_required_config_is_rejected(capsys) -> None:
    code = main(["--key-provider", "vault", "config", "show"])
    assert code == 2
    err = capsys.readouterr().err
    assert "config_invalid_key_provider" in err


def test_vault_provider_mtls_without_local_cert_files_is_allowed(tmp_path: Path, capsys) -> None:
    config_path = tmp_path / "config.json"
    config_path.write_text(
        json.dumps(
            {
                "storage_dir": str(tmp_path / "data"),
                "key_provider": "vault",
                "vault_url": "https://vault.example",
                "vault_path": "secret/data/acp/identities",
                "mtls_enabled": True,
            },
        ),
        encoding="utf-8",
    )

    code = main(["--config", str(config_path), "--json", "config", "validate"])
    assert code == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["ok"] is True


def test_vault_provider_mtls_rejects_partial_local_cert_override(tmp_path: Path, capsys) -> None:
    config_path = tmp_path / "config.json"
    config_path.write_text(
        json.dumps(
            {
                "storage_dir": str(tmp_path / "data"),
                "key_provider": "vault",
                "vault_url": "https://vault.example",
                "vault_path": "secret/data/acp/identities",
                "mtls_enabled": True,
                "cert_file": "/tmp/client-cert.pem",
            },
        ),
        encoding="utf-8",
    )

    code = main(["--config", str(config_path), "--json", "config", "show"])
    assert code == 2
    err = capsys.readouterr().err
    assert "config_invalid_http_security" in err



# Copyright 2026 ACP Project
# Licensed under the Apache License, Version 2.0
# See LICENSE file for details.

from __future__ import annotations

from dataclasses import dataclass
from functools import wraps
from typing import Any, Callable

from .agent import Agent
from .messages import FailReason
from .overlay import (
    OverlayAdapterError,
    OverlayInboundAdapter,
    OverlayOutboundAdapter,
)


BusinessHandler = Callable[[dict[str, Any]], dict[str, Any] | None]
PassThroughHandler = Callable[[dict[str, Any]], dict[str, Any] | None]
InboundDecoratorHandler = Callable[[dict[str, Any]], dict[str, Any] | None]

WELL_KNOWN_CACHE_CONTROL = "public, max-age=300"

# Runtime alias populated in register_fastapi_overlay_routes when FastAPI is available.
_FastApiRequest = Any


class OverlayFrameworkError(RuntimeError):
    pass


@dataclass(frozen=True)
class OverlayHttpResponse:
    status_code: int
    body: dict[str, Any]


@dataclass
class OverlayFrameworkRuntime:
    agent: Agent
    base_url: str
    inbound_adapter: OverlayInboundAdapter
    outbound_adapter: OverlayOutboundAdapter

    @classmethod
    def create(
        cls,
        *,
        agent: Agent,
        base_url: str,
        business_handler: BusinessHandler,
        passthrough_handler: PassThroughHandler | None = None,
    ) -> "OverlayFrameworkRuntime":
        normalized_base_url = str(base_url or "").strip()
        if not normalized_base_url:
            raise OverlayFrameworkError("base_url is required")
        return cls(
            agent=agent,
            base_url=normalized_base_url.rstrip("/"),
            inbound_adapter=OverlayInboundAdapter(
                agent=agent,
                business_handler=business_handler,
                passthrough_handler=passthrough_handler,
            ),
            outbound_adapter=OverlayOutboundAdapter(agent),
        )

    def handle_message_body(self, body: Any) -> OverlayHttpResponse:
        if not isinstance(body, dict):
            return OverlayHttpResponse(
                status_code=400,
                body={
                    "mode": "invalid",
                    "state": "FAILED",
                    "reason_code": FailReason.POLICY_REJECTED.value,
                    "detail": "Expected JSON object request body",
                    "response_message": None,
                },
            )
        try:
            result = self.inbound_adapter.handle_request(body)
        except OverlayAdapterError as exc:
            return OverlayHttpResponse(
                status_code=400,
                body={
                    "mode": "invalid",
                    "state": "FAILED",
                    "reason_code": FailReason.POLICY_REJECTED.value,
                    "detail": str(exc),
                    "response_message": None,
                },
            )
        return OverlayHttpResponse(status_code=200, body=result)

    def well_known_document(self) -> dict[str, Any]:
        return self.agent.build_well_known_document(base_url=self.base_url)

    def identity_document_payload(self) -> dict[str, Any]:
        return {"identity_document": self.agent.identity_document}

    @staticmethod
    def well_known_headers() -> dict[str, str]:
        return {"Cache-Control": WELL_KNOWN_CACHE_CONTROL}

    def send_business_payload(
        self,
        *,
        payload: dict[str, Any],
        target_base_url: str | None = None,
        recipient_agent_id: str | None = None,
        context: str | None = None,
        delivery_mode: str = "auto",
        expires_in_seconds: int = 300,
    ) -> dict[str, Any]:
        target, send_result = self.outbound_adapter.send_business_payload(
            payload=payload,
            target_base_url=target_base_url,
            recipient_agent_id=recipient_agent_id,
            context=context,
            delivery_mode=delivery_mode,
            expires_in_seconds=expires_in_seconds,
        )
        return {
            "target": (
                {
                    "agent_id": target.agent_id,
                    "base_url": target.base_url,
                    "well_known_url": target.well_known_url,
                    "identity_document_url": target.identity_document_url,
                }
                if target is not None
                else None
            ),
            "send_result": send_result.to_dict(),
        }

    def send_acp(
        self,
        target_url: str,
        payload: dict[str, Any],
        *,
        recipient_agent_id: str | None = None,
        context: str | None = None,
        delivery_mode: str = "auto",
        expires_in_seconds: int = 300,
    ) -> dict[str, Any]:
        return self.send_business_payload(
            payload=payload,
            target_base_url=target_url,
            recipient_agent_id=recipient_agent_id,
            context=context,
            delivery_mode=delivery_mode,
            expires_in_seconds=expires_in_seconds,
        )


@dataclass
class OverlayClient:
    agent: Agent
    outbound_adapter: OverlayOutboundAdapter

    @classmethod
    def create(cls, *, agent: Agent) -> "OverlayClient":
        return cls(agent=agent, outbound_adapter=OverlayOutboundAdapter(agent))

    def send_acp(
        self,
        target_url: str,
        payload: dict[str, Any],
        *,
        recipient_agent_id: str | None = None,
        context: str | None = None,
        delivery_mode: str = "auto",
        expires_in_seconds: int = 300,
    ) -> dict[str, Any]:
        target, send_result = self.outbound_adapter.send_business_payload(
            payload=payload,
            target_base_url=target_url,
            recipient_agent_id=recipient_agent_id,
            context=context,
            delivery_mode=delivery_mode,
            expires_in_seconds=expires_in_seconds,
        )
        return {
            "target": (
                {
                    "agent_id": target.agent_id,
                    "base_url": target.base_url,
                    "well_known_url": target.well_known_url,
                    "identity_document_url": target.identity_document_url,
                }
                if target is not None
                else None
            ),
            "send_result": send_result.to_dict(),
        }


def _agent_from_overlay_config(config: Any) -> Agent:
    if isinstance(config, Agent):
        return config
    if isinstance(config, OverlayFrameworkRuntime):
        return config.agent
    if isinstance(config, dict):
        candidate = config.get("agent")
        if isinstance(candidate, Agent):
            return candidate
    raise OverlayFrameworkError("acp_overlay_inbound requires config to provide an Agent")


def acp_overlay_inbound(
    config: Any,
    *,
    passthrough: bool = True,
) -> Callable[[InboundDecoratorHandler], Callable[[dict[str, Any]], dict[str, Any]]]:
    agent = _agent_from_overlay_config(config)

    def decorator(handler: InboundDecoratorHandler) -> Callable[[dict[str, Any]], dict[str, Any]]:
        if not callable(handler):
            raise OverlayFrameworkError("acp_overlay_inbound requires a callable handler")
        passthrough_handler: PassThroughHandler | None = None
        if passthrough:
            passthrough_handler = lambda body: handler(body)
        inbound = OverlayInboundAdapter(
            agent=agent,
            business_handler=handler,
            passthrough_handler=passthrough_handler,
        )

        @wraps(handler)
        def wrapper(payload: dict[str, Any]) -> dict[str, Any]:
            if not isinstance(payload, dict):
                raise OverlayFrameworkError(
                    "acp_overlay_inbound wrapped handlers require a JSON object payload argument",
                )
            return inbound.handle_request(payload)

        return wrapper

    return decorator


def register_fastapi_overlay_routes(
    app: Any,
    *,
    runtime: OverlayFrameworkRuntime,
    message_path: str = "/api/v1/acp/messages",
    well_known_path: str = "/.well-known/acp",
    identity_path: str = "/api/v1/acp/identity",
) -> None:
    try:
        from fastapi import Request
        from fastapi.responses import JSONResponse
    except ImportError as exc:  # pragma: no cover - optional runtime dependency
        raise OverlayFrameworkError("FastAPI is not installed; cannot register FastAPI overlay routes") from exc

    globals()["_FastApiRequest"] = Request

    @app.post(message_path)
    async def _acp_overlay_message(request: _FastApiRequest) -> Any:
        try:
            body = await request.json()
        except Exception:  # pragma: no cover - framework parser behavior
            body = None
        response = runtime.handle_message_body(body)
        return JSONResponse(status_code=response.status_code, content=response.body)

    @app.get(well_known_path)
    async def _acp_overlay_well_known() -> Any:
        return JSONResponse(
            status_code=200,
            content=runtime.well_known_document(),
            headers=runtime.well_known_headers(),
        )

    @app.get(identity_path)
    async def _acp_overlay_identity() -> dict[str, Any]:
        return runtime.identity_document_payload()


def register_flask_overlay_routes(
    app: Any,
    *,
    runtime: OverlayFrameworkRuntime,
    message_path: str = "/api/v1/acp/messages",
    well_known_path: str = "/.well-known/acp",
    identity_path: str = "/api/v1/acp/identity",
) -> None:
    try:
        from flask import jsonify, request
    except ImportError as exc:  # pragma: no cover - optional runtime dependency
        raise OverlayFrameworkError("Flask is not installed; cannot register Flask overlay routes") from exc

    def _message() -> Any:
        body = request.get_json(silent=True)
        response = runtime.handle_message_body(body)
        return jsonify(response.body), response.status_code

    def _well_known() -> Any:
        response = jsonify(runtime.well_known_document())
        response.headers["Cache-Control"] = runtime.well_known_headers()["Cache-Control"]
        return response

    def _identity() -> Any:
        return jsonify(runtime.identity_document_payload())

    app.add_url_rule(message_path, endpoint="acp_overlay_message", view_func=_message, methods=["POST"])
    app.add_url_rule(well_known_path, endpoint="acp_overlay_well_known", view_func=_well_known, methods=["GET"])
    app.add_url_rule(identity_path, endpoint="acp_overlay_identity", view_func=_identity, methods=["GET"])



from __future__ import annotations

from pathlib import Path
from typing import Any

import pytest
import requests

from acp import Agent, DeliveryState
from acp.transport import HTTPTransport, TransportError


class DummyResponse:
    def __init__(self, status_code: int, body: dict[str, Any] | None = None) -> None:
        self.status_code = status_code
        self._body = body
        self.text = "" if body is None else str(body)

    def json(self) -> dict[str, Any]:
        if self._body is None:
            raise ValueError("No JSON body")
        return self._body


def test_http_transport_retries_on_transient_network_error(monkeypatch: pytest.MonkeyPatch) -> None:
    calls: list[int] = []

    def fake_request(method: str, url: str, **_: Any) -> DummyResponse:
        calls.append(1)
        if len(calls) < 3:
            raise requests.RequestException("transient failure")
        return DummyResponse(200, {"status": "ok"})

    monkeypatch.setattr(requests, "request", fake_request)
    transport = HTTPTransport(
        max_retries=2,
        retry_backoff_seconds=0.0,
        allow_insecure_http=True,
    )
    response = transport.post_json("http://localhost:9000/test", {"hello": "world"})

    assert response.status_code == 200
    assert len(calls) == 3


def test_http_transport_raises_after_retry_exhaustion(monkeypatch: pytest.MonkeyPatch) -> None:
    def fake_request(method: str, url: str, **_: Any) -> DummyResponse:
        raise requests.RequestException("still failing")

    monkeypatch.setattr(requests, "request", fake_request)
    transport = HTTPTransport(
        max_retries=1,
        retry_backoff_seconds=0.0,
        allow_insecure_http=True,
    )

    with pytest.raises(TransportError):
        transport.post_json("http://localhost:9000/test", {"hello": "world"})


def test_send_direct_delivery_mode_uses_recipient_endpoint(tmp_path: Path) -> None:
    sender = Agent.create(
        "agent:sender.bot@localhost:9300",
        storage_dir=tmp_path / "sender",
        endpoint="http://localhost:9300/acp/inbox",
        relay_url="http://localhost:8080",
        relay_hints=["http://localhost:8080"],
        discovery_scheme="http",
        allow_insecure_http=True,
    )
    recipient = Agent.create(
        "agent:recipient.bot@localhost:9301",
        storage_dir=tmp_path / "recipient",
        endpoint="http://localhost:9301/acp/inbox",
        relay_url="http://localhost:8080",
        relay_hints=["http://localhost:8080"],
        discovery_scheme="http",
        allow_insecure_http=True,
    )

    class InMemoryDirectTransport:
        def __init__(self) -> None:
            self.urls: list[str] = []

        def post_json(self, url: str, body: dict[str, Any]) -> DummyResponse:
            self.urls.append(url)
            if url != "http://localhost:9301/acp/inbox":
                return DummyResponse(404, {"detail": "unknown endpoint"})
            result = recipient.handle_incoming(body)
            return DummyResponse(200, result)

    direct_transport = InMemoryDirectTransport()
    sender.relay_client.transport = direct_transport  # type: ignore[assignment]
    sender.discovery.resolve = (  # type: ignore[method-assign]
        lambda agent_id: recipient.identity_document
        if agent_id == recipient.agent_id
        else sender.identity_document
    )

    result = sender.send(
        recipients=[recipient.agent_id],
        payload={"type": "ping"},
        context="ctx-direct",
        delivery_mode="direct",
    )

    assert direct_transport.urls == ["http://localhost:9301/acp/inbox"]
    assert result.outcomes[0].state is DeliveryState.ACKNOWLEDGED
    assert result.message_ids == [result.message_id]



from __future__ import annotations

import json
from pathlib import Path
from typing import Any

import pytest
import requests

from acp.capabilities import AgentCapabilities
from acp.discovery import DiscoveryClient
from acp.identity import AgentIdentity, read_identity
from acp_cli.main import build_parser, main


class DummyResponse:
    def __init__(self, status_code: int, body: dict[str, Any] | None = None) -> None:
        self.status_code = status_code
        self._body = body

    def json(self) -> dict[str, Any]:
        if self._body is None:
            raise ValueError("No JSON body")
        return self._body


def test_parser_accepts_identity_and_discover_commands() -> None:
    parser = build_parser()
    args = parser.parse_args(["identity", "create", "--agent-id", "agent:alice@localhost:9001"])
    assert args.domain == "identity"
    assert args.identity_command == "create"

    args = parser.parse_args(["discover", "get", "--agent-id", "agent:bob@localhost:9002"])
    assert args.domain == "discover"
    assert args.discover_command == "get"

    args = parser.parse_args(["discover", "well-known", "https://agent.example"])
    assert args.domain == "discover"
    assert args.discover_command == "well-known"


def test_version_flag_outputs_version(capsys) -> None:
    code = main(["--version"])
    assert code == 0
    out = capsys.readouterr().out.strip()
    assert out.startswith("acp ")


def test_identity_create_and_show_json(tmp_path: Path, capsys) -> None:
    agent_id = "agent:test.identity@localhost:9100"
    code = main(["--storage-dir", str(tmp_path), "identity", "create", "--agent-id", agent_id])
    assert code == 0
    capsys.readouterr()

    assert read_identity(tmp_path, agent_id) is not None

    code = main(["--storage-dir", str(tmp_path), "--json", "identity", "show", "--agent-id", agent_id])
    assert code == 0
    out = capsys.readouterr().out
    payload = json.loads(out)
    assert payload["ok"] is True
    assert payload["agent_id"] == agent_id
    assert "public_keys" in payload
    dumped = json.dumps(payload)
    assert "signing_private_key" not in dumped
    assert "encryption_private_key" not in dumped


def test_identity_export_and_verify(tmp_path: Path, capsys) -> None:
    agent_id = "agent:test.export@localhost:9200"
    export_path = tmp_path / "exported" / "identity_document.json"

    assert main(["--storage-dir", str(tmp_path), "identity", "create", "--agent-id", agent_id]) == 0
    capsys.readouterr()

    assert main(
        [
            "--storage-dir",
            str(tmp_path),
            "identity",
            "export",
            "--agent-id",
            agent_id,
            "--out",
            str(export_path),
        ],
    ) == 0
    capsys.readouterr()
    assert export_path.exists()

    code = main(["--json", "identity", "verify", "--file", str(export_path)])
    assert code == 0
    verified = json.loads(capsys.readouterr().out)
    assert verified["valid"] is True

    tampered = json.loads(export_path.read_text(encoding="utf-8"))
    tampered["trust_profile"] = "invalid_profile"
    export_path.write_text(json.dumps(tampered), encoding="utf-8")
    code = main(["--json", "identity", "verify", "--file", str(export_path)])
    assert code == 1
    verified = json.loads(capsys.readouterr().out)
    assert verified["valid"] is False


def test_discover_get_and_list_from_cache(tmp_path: Path, capsys) -> None:
    target_id = "agent:discover.target"
    identity = AgentIdentity.create(target_id)
    identity_document = identity.build_identity_document(
        direct_endpoint="http://localhost:9999/api/v1/acp/messages",
        relay_hints=["http://localhost:8080"],
        trust_profile="self_asserted",
        capabilities=AgentCapabilities(agent_id=target_id).to_dict(),
    )

    cache_path = tmp_path / "discovery_cache.json"
    discovery = DiscoveryClient(cache_path=cache_path)
    discovery.seed(identity_document)

    code = main(["--storage-dir", str(tmp_path), "--json", "discover", "get", "--agent-id", target_id])
    assert code == 0
    get_payload = json.loads(capsys.readouterr().out)
    assert get_payload["ok"] is True
    assert get_payload["resolved"]["agent_id"] == target_id
    assert get_payload["resolved"]["service"]["direct_endpoint"] == "http://localhost:9999/api/v1/acp/messages"

    code = main(["--storage-dir", str(tmp_path), "--json", "discover", "list"])
    assert code == 0
    list_payload = json.loads(capsys.readouterr().out)
    assert list_payload["ok"] is True
    assert list_payload["count"] == 1
    assert list_payload["entries"][0]["agent_id"] == target_id


def test_discover_well_known_command(
    tmp_path: Path,
    capsys,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    target_id = "agent:wellknown.target@company.local"
    identity = AgentIdentity.create(target_id)
    identity_document = identity.build_identity_document(
        direct_endpoint="https://company.local/api/v1/acp/messages",
        relay_hints=["https://relay.company.local"],
        trust_profile="self_asserted",
        capabilities=AgentCapabilities(agent_id=target_id).to_dict(),
    )
    well_known = {
        "agent_id": target_id,
        "identity_document": "https://company.local/api/v1/acp/identity",
        "transports": {"http": {"endpoint": "https://company.local/api/v1/acp/messages"}},
        "version": "1.0",
        "security_profile": "https",
    }

    def fake_get(
        url: str,
        params: dict[str, str] | None = None,
        timeout: int = 5,
        **_kwargs: object,
    ) -> DummyResponse:
        if params is not None:
            return DummyResponse(404)
        if url == "https://company.local/.well-known/acp":
            return DummyResponse(200, well_known)
        if url == "https://company.local/api/v1/acp/identity":
            return DummyResponse(200, {"identity_document": identity_document})
        return DummyResponse(404)

    monkeypatch.setattr(requests, "get", fake_get)

    code = main(
        [
            "--storage-dir",
            str(tmp_path),
            "--json",
            "discover",
            "well-known",
            "https://company.local",
            "--agent-id",
            target_id,
        ],
    )
    assert code == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["ok"] is True
    assert payload["well_known_url"] == "https://company.local/.well-known/acp"
    assert payload["well_known"]["agent_id"] == target_id
    assert payload["resolved"]["agent_id"] == target_id


def test_missing_argument_and_bad_argument_handling(tmp_path: Path, capsys) -> None:
    code = main(["identity", "create"])
    assert code == 2
    err = capsys.readouterr().err
    assert "required" in err.lower()

    code = main(
        [
            "--storage-dir",
            str(tmp_path),
            "identity",
            "create",
            "--agent-id",
            "agent:test.bad@localhost:9300",
            "--trust-profile",
            "invalid_profile",
        ],
    )
    assert code == 2
    err = capsys.readouterr().err
    assert "identity_create_failed" in err



from acp.crypto import decrypt_for_recipient, encrypt_for_recipients, sign_protected_payload, verify_protected_payload_signature
from acp.identity import AgentIdentity
from acp.messages import Envelope, MessageClass


def test_encrypt_sign_verify_decrypt_roundtrip() -> None:
    sender = AgentIdentity.create("agent:sender@localhost:9101")
    recipient = AgentIdentity.create("agent:recipient@localhost:9102")

    envelope = Envelope.build(
        sender=sender.agent_id,
        recipients=[recipient.agent_id],
        message_class=MessageClass.SEND,
        context_id="ctx-1",
        expires_in_seconds=60,
    )
    payload = {"type": "hello", "data": {"value": 42}}

    protected = encrypt_for_recipients(
        payload,
        envelope,
        {recipient.agent_id: recipient.encryption_public_key},
    )
    protected = sign_protected_payload(
        envelope,
        protected,
        sender.signing_private_key,
        sender.signing_kid,
    )

    assert verify_protected_payload_signature(
        envelope,
        protected,
        sender.signing_public_key,
    )
    decrypted = decrypt_for_recipient(
        envelope,
        protected,
        recipient.agent_id,
        recipient.encryption_private_key,
    )
    assert decrypted == payload



# Copyright 2026 ACP Project
# Licensed under the Apache License, Version 2.0
# See LICENSE file for details.

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable

from .agent import Agent
from .messages import SendResult


class OverlayAdapterError(RuntimeError):
    pass


BusinessHandler = Callable[[dict[str, Any]], dict[str, Any] | None]
PassThroughHandler = Callable[[dict[str, Any]], dict[str, Any] | None]


def is_acp_http_message(body: Any) -> bool:
    if not isinstance(body, dict):
        return False
    envelope = body.get("envelope")
    protected = body.get("protected")
    return isinstance(envelope, dict) and isinstance(protected, dict)


@dataclass(frozen=True)
class OverlayTarget:
    agent_id: str
    base_url: str
    well_known_url: str
    identity_document_url: str


@dataclass
class OverlayInboundAdapter:
    agent: Agent
    business_handler: BusinessHandler
    passthrough_handler: PassThroughHandler | None = None

    def handle_request(self, body: dict[str, Any]) -> dict[str, Any]:
        if not isinstance(body, dict):
            raise OverlayAdapterError("Overlay inbound adapter requires a JSON object request body")
        if not is_acp_http_message(body):
            if self.passthrough_handler is None:
                raise OverlayAdapterError("Request is not an ACP message and no passthrough_handler is configured")
            payload = self.passthrough_handler(body)
            return {
                "mode": "passthrough",
                "payload": payload if isinstance(payload, dict) else None,
            }
        acp_result = self.agent.handle_incoming(
            body,
            handler=lambda payload, _envelope: self.business_handler(payload),
        )
        return {
            "mode": "acp",
            "acp_result": acp_result,
            "state": acp_result.get("state"),
            "reason_code": acp_result.get("reason_code"),
            "detail": acp_result.get("detail"),
            "response_message": acp_result.get("response_message"),
        }


@dataclass
class OverlayOutboundAdapter:
    agent: Agent

    def resolve_target(
        self,
        *,
        target_base_url: str,
        expected_agent_id: str | None = None,
    ) -> OverlayTarget:
        resolved = self.agent.discovery.resolve_well_known(
            target_base_url,
            expected_agent_id=expected_agent_id,
        )
        well_known = resolved["well_known"]
        identity_document = resolved["identity_document"]
        agent_id = identity_document.get("agent_id")
        if not isinstance(agent_id, str) or not agent_id.strip():
            raise OverlayAdapterError("Resolved well-known metadata did not include a valid identity_document.agent_id")
        identity_document_url = well_known.get("identity_document")
        if not isinstance(identity_document_url, str) or not identity_document_url.strip():
            raise OverlayAdapterError("Resolved well-known metadata did not include a valid identity_document URL")
        return OverlayTarget(
            agent_id=agent_id,
            base_url=target_base_url.rstrip("/"),
            well_known_url=resolved["well_known_url"],
            identity_document_url=identity_document_url,
        )

    def send_business_payload(
        self,
        *,
        payload: dict[str, Any],
        target_base_url: str | None = None,
        recipient_agent_id: str | None = None,
        context: str | None = None,
        delivery_mode: str = "auto",
        expires_in_seconds: int = 300,
    ) -> tuple[OverlayTarget | None, SendResult]:
        resolved_target: OverlayTarget | None = None
        resolved_recipient = recipient_agent_id
        if target_base_url:
            resolved_target = self.resolve_target(
                target_base_url=target_base_url,
                expected_agent_id=recipient_agent_id,
            )
            if resolved_recipient is None:
                resolved_recipient = resolved_target.agent_id
        if not isinstance(resolved_recipient, str) or not resolved_recipient.strip():
            raise OverlayAdapterError(
                "send_business_payload requires recipient_agent_id or target_base_url for well-known bootstrap",
            )
        send_result = self.agent.send(
            recipients=[resolved_recipient],
            payload=payload,
            context=context,
            delivery_mode=delivery_mode,
            expires_in_seconds=expires_in_seconds,
        )
        return resolved_target, send_result



# Copyright 2026 ACP Project
# Licensed under the Apache License, Version 2.0
# See LICENSE file for details.

from .agent import Agent, ProcessingError
from .amqp_transport import (
    AMQPTransport,
    AMQPTransportError,
    build_amqp_service_hint,
    queue_name_for_agent,
    routing_key_for_agent,
)
from .mqtt_transport import (
    DEFAULT_MQTT_QOS,
    DEFAULT_MQTT_TOPIC_PREFIX,
    MQTTTransport,
    MQTTTransportError,
    build_mqtt_service_hint,
    topic_for_agent,
)
from .overlay import (
    OverlayAdapterError,
    OverlayInboundAdapter,
    OverlayOutboundAdapter,
    OverlayTarget,
    is_acp_http_message,
)
from .overlay_framework import (
    WELL_KNOWN_CACHE_CONTROL,
    OverlayClient,
    OverlayFrameworkError,
    OverlayFrameworkRuntime,
    OverlayHttpResponse,
    acp_overlay_inbound,
    register_fastapi_overlay_routes,
    register_flask_overlay_routes,
)
from .well_known import (
    DEFAULT_IDENTITY_DOCUMENT_PATH,
    WELL_KNOWN_PATH,
    build_well_known_document,
    well_known_url_from_base,
)
from .capabilities import AgentCapabilities, CapabilityMatch, choose_compatible
from .key_provider import IdentityKeyMaterial, KeyProviderError, LocalKeyProvider, TlsMaterial, VaultKeyProvider
from .messages import (
    ACPMessage,
    CompensateInstruction,
    DeliveryOutcome,
    DeliveryState,
    Envelope,
    FailReason,
    MessageClass,
    ProtectedPayload,
    SendResult,
    WrappedContentKey,
)

__all__ = [
    "ACPMessage",
    "AMQPTransport",
    "AMQPTransportError",
    "Agent",
    "AgentCapabilities",
    "CapabilityMatch",
    "CompensateInstruction",
    "DeliveryOutcome",
    "DeliveryState",
    "Envelope",
    "FailReason",
    "MessageClass",
    "MQTTTransport",
    "MQTTTransportError",
    "OverlayAdapterError",
    "OverlayClient",
    "OverlayInboundAdapter",
    "OverlayOutboundAdapter",
    "OverlayTarget",
    "OverlayFrameworkError",
    "OverlayFrameworkRuntime",
    "OverlayHttpResponse",
    "acp_overlay_inbound",
    "WELL_KNOWN_CACHE_CONTROL",
    "LocalKeyProvider",
    "VaultKeyProvider",
    "IdentityKeyMaterial",
    "TlsMaterial",
    "KeyProviderError",
    "ProcessingError",
    "ProtectedPayload",
    "SendResult",
    "WrappedContentKey",
    "build_amqp_service_hint",
    "build_mqtt_service_hint",
    "DEFAULT_MQTT_QOS",
    "DEFAULT_MQTT_TOPIC_PREFIX",
    "DEFAULT_IDENTITY_DOCUMENT_PATH",
    "WELL_KNOWN_PATH",
    "build_well_known_document",
    "is_acp_http_message",
    "register_fastapi_overlay_routes",
    "register_flask_overlay_routes",
    "queue_name_for_agent",
    "routing_key_for_agent",
    "topic_for_agent",
    "well_known_url_from_base",
    "choose_compatible",
]



# Copyright 2026 ACP Project
# Licensed under the Apache License, Version 2.0
# See LICENSE file for details.

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from pathlib import Path
import re
from typing import Any
import uuid

from .crypto import canonical_json, generate_ed25519_keypair, generate_x25519_keypair, sign_bytes, verify_signature


IDENTITY_FILE_NAME = "identity.json"
IDENTITY_DOC_FILE_NAME = "identity_document.json"
_AGENT_ID_PATTERN = re.compile(r"^agent:(?P<name>[^@]+)(?:@(?P<domain>.+))?$")
TRUST_PROFILES = {
    "self_asserted",
    "domain_verified",
    "enterprise_managed",
    "regulated_assured",
}


class IdentityError(ValueError):
    pass


def parse_agent_id(agent_id: str) -> tuple[str, str | None]:
    match = _AGENT_ID_PATTERN.match(agent_id)
    if match is None:
        raise IdentityError(f"Invalid agent identifier: {agent_id}")
    return match.group("name"), match.group("domain")


def sanitize_agent_id(agent_id: str) -> str:
    return re.sub(r"[^a-zA-Z0-9._-]", "_", agent_id)


def identity_path(storage_dir: Path, agent_id: str) -> Path:
    return storage_dir / sanitize_agent_id(agent_id)


def _utc_now_iso() -> str:
    return datetime.now(timezone.utc).isoformat().replace("+00:00", "Z")


@dataclass
class AgentIdentity:
    agent_id: str
    signing_private_key: str
    signing_public_key: str
    encryption_private_key: str
    encryption_public_key: str
    signing_kid: str
    encryption_kid: str

    @classmethod
    def create(cls, agent_id: str) -> "AgentIdentity":
        parse_agent_id(agent_id)
        signing_private_key, signing_public_key = generate_ed25519_keypair()
        encryption_private_key, encryption_public_key = generate_x25519_keypair()
        return cls(
            agent_id=agent_id,
            signing_private_key=signing_private_key,
            signing_public_key=signing_public_key,
            encryption_private_key=encryption_private_key,
            encryption_public_key=encryption_public_key,
            signing_kid=f"sig-{uuid.uuid4().hex[:12]}",
            encryption_kid=f"enc-{uuid.uuid4().hex[:12]}",
        )

    def to_dict(self) -> dict[str, str]:
        return {
            "agent_id": self.agent_id,
            "signing_private_key": self.signing_private_key,
            "signing_public_key": self.signing_public_key,
            "encryption_private_key": self.encryption_private_key,
            "encryption_public_key": self.encryption_public_key,
            "signing_kid": self.signing_kid,
            "encryption_kid": self.encryption_kid,
        }

    @classmethod
    def from_dict(cls, value: dict[str, Any]) -> "AgentIdentity":
        return cls(
            agent_id=str(value["agent_id"]),
            signing_private_key=str(value["signing_private_key"]),
            signing_public_key=str(value["signing_public_key"]),
            encryption_private_key=str(value["encryption_private_key"]),
            encryption_public_key=str(value["encryption_public_key"]),
            signing_kid=str(value["signing_kid"]),
            encryption_kid=str(value["encryption_kid"]),
        )

    def build_identity_document(
        self,
        *,
        direct_endpoint: str | None,
        relay_hints: list[str] | None,
        http_security_profile: str | None = None,
        relay_security_profile: str | None = None,
        amqp_service: dict[str, Any] | None = None,
        mqtt_service: dict[str, Any] | None = None,
        trust_profile: str,
        capabilities: dict[str, Any] | None = None,
        valid_days: int = 365,
    ) -> dict[str, Any]:
        if trust_profile not in TRUST_PROFILES:
            raise IdentityError(
                f"Unsupported trust_profile {trust_profile}; expected one of {sorted(TRUST_PROFILES)}",
            )
        service: dict[str, Any] = {
            "direct_endpoint": direct_endpoint,
            "relay_hints": relay_hints or [],
        }
        if isinstance(direct_endpoint, str) and direct_endpoint and http_security_profile:
            service["http"] = {
                "endpoint": direct_endpoint,
                "security_profile": str(http_security_profile),
            }
        if relay_hints and relay_security_profile:
            service["relay"] = {
                "endpoint": str(relay_hints[0]),
                "security_profile": str(relay_security_profile),
            }
        if amqp_service:
            service["amqp"] = dict(amqp_service)
        if mqtt_service:
            service["mqtt"] = dict(mqtt_service)

        document = {
            "acp_identity_version": "1.0",
            "agent_id": self.agent_id,
            "created_at": _utc_now_iso(),
            "valid_until": (
                datetime.now(timezone.utc) + timedelta(days=valid_days)
            ).isoformat().replace("+00:00", "Z"),
            "trust_profile": trust_profile,
            "keys": {
                "signing": {
                    "kid": self.signing_kid,
                    "alg": "Ed25519",
                    "public_key": self.signing_public_key,
                },
                "encryption": {
                    "kid": self.encryption_kid,
                    "alg": "X25519",
                    "public_key": self.encryption_public_key,
                },
            },
            "service": service,
            "capabilities": capabilities or {},
        }
        to_sign = canonical_json(document).encode("utf-8")
        document["signature"] = {
            "algorithm": "Ed25519",
            "signed_by": self.signing_kid,
            "value": sign_bytes(to_sign, self.signing_private_key),
        }
        return document


def verify_identity_document(identity_document: dict[str, Any]) -> bool:
    required = ["agent_id", "keys", "service", "signature", "valid_until"]
    for key in required:
        if key not in identity_document:
            return False

    signature = identity_document.get("signature", {})
    signature_value = signature.get("value")
    signing_public_key = (
        identity_document.get("keys", {})
        .get("signing", {})
        .get("public_key")
    )
    if not signature_value or not signing_public_key:
        return False

    trust_profile = identity_document.get("trust_profile")
    if trust_profile not in TRUST_PROFILES:
        return False

    valid_until = identity_document.get("valid_until")
    if not isinstance(valid_until, str):
        return False
    try:
        if datetime.fromisoformat(valid_until.replace("Z", "+00:00")) <= datetime.now(
            timezone.utc,
        ):
            return False
    except Exception:  # noqa: BLE001
        return False

    unsigned_document = dict(identity_document)
    unsigned_document.pop("signature", None)
    to_verify = canonical_json(unsigned_document).encode("utf-8")
    return verify_signature(to_verify, signature_value, signing_public_key)


def write_identity(
    storage_dir: Path,
    identity: AgentIdentity,
    identity_document: dict[str, Any],
) -> None:
    agent_dir = identity_path(storage_dir, identity.agent_id)
    agent_dir.mkdir(parents=True, exist_ok=True)
    identity_file = agent_dir / IDENTITY_FILE_NAME
    document_file = agent_dir / IDENTITY_DOC_FILE_NAME
    identity_file.write_text(canonical_json(identity.to_dict()), encoding="utf-8")
    document_file.write_text(canonical_json(identity_document), encoding="utf-8")


def read_identity(storage_dir: Path, agent_id: str) -> tuple[AgentIdentity, dict[str, Any]] | None:
    agent_dir = identity_path(storage_dir, agent_id)
    identity_file = agent_dir / IDENTITY_FILE_NAME
    document_file = agent_dir / IDENTITY_DOC_FILE_NAME
    if not identity_file.exists() or not document_file.exists():
        return None

    import json

    identity = AgentIdentity.from_dict(json.loads(identity_file.read_text(encoding="utf-8")))
    identity_document = json.loads(document_file.read_text(encoding="utf-8"))
    return identity, identity_document



# Copyright 2026 ACP Project
# Licensed under the Apache License, Version 2.0
# See LICENSE file for details.

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Optional
import uuid
import warnings
from urllib.parse import urlsplit

from .amqp_transport import (
    AMQPTransport,
    AMQPTransportError,
    DEFAULT_AMQP_EXCHANGE,
    build_amqp_service_hint,
)
from .mqtt_transport import (
    DEFAULT_MQTT_QOS,
    DEFAULT_MQTT_TOPIC_PREFIX,
    MQTTTransport,
    MQTTTransportError,
    build_mqtt_service_hint,
)
from .capabilities import AgentCapabilities, choose_compatible
from .crypto import (
    CryptoError,
    decrypt_for_recipient,
    encrypt_for_recipients,
    sign_protected_payload,
    verify_protected_payload_signature,
)
from .discovery import DiscoveryClient, DiscoveryError
from .http_security import (
    HttpSecurityError,
    HttpSecurityPolicy,
    enforce_http_security,
    validate_http_security_policy,
)
from .identity import AgentIdentity, parse_agent_id, read_identity, verify_identity_document, write_identity
from .key_provider import IdentityKeyMaterial, KeyProvider, KeyProviderError, LocalKeyProvider
from .messages import (
    ACP_VERSION,
    ACPMessage,
    CompensateInstruction,
    DEFAULT_CRYPTO_SUITE,
    DeliveryOutcome,
    DeliveryState,
    Envelope,
    FailReason,
    MessageClass,
    SendResult,
    build_ack_payload,
    build_fail_payload,
)
from .relay_client import RelayClient
from .transport import TransportError
from .well_known import build_well_known_document as build_well_known_metadata


IncomingHandler = Callable[[dict[str, Any], Envelope], Optional[dict[str, Any]]]


@dataclass
class ProcessingError(RuntimeError):
    reason_code: FailReason
    detail: str


def _identity_from_provider(agent_id: str, keys: IdentityKeyMaterial) -> AgentIdentity:
    missing: list[str] = []
    if not keys.signing_public_key:
        missing.append("signing_public_key")
    if not keys.encryption_public_key:
        missing.append("encryption_public_key")
    if not keys.signing_kid:
        missing.append("signing_kid")
    if not keys.encryption_kid:
        missing.append("encryption_kid")
    if missing:
        raise TransportError(
            "External key provider requires identity public metadata for first-time bootstrap: "
            + ", ".join(missing),
        )
    return AgentIdentity(
        agent_id=agent_id,
        signing_private_key=keys.signing_private_key,
        signing_public_key=str(keys.signing_public_key),
        encryption_private_key=keys.encryption_private_key,
        encryption_public_key=str(keys.encryption_public_key),
        signing_kid=str(keys.signing_kid),
        encryption_kid=str(keys.encryption_kid),
    )


def _apply_provider_keys(identity: AgentIdentity, keys: IdentityKeyMaterial) -> AgentIdentity:
    if keys.signing_public_key and keys.signing_public_key != identity.signing_public_key:
        raise TransportError("Key provider signing_public_key does not match local identity metadata")
    if keys.encryption_public_key and keys.encryption_public_key != identity.encryption_public_key:
        raise TransportError("Key provider encryption_public_key does not match local identity metadata")
    if keys.signing_kid and keys.signing_kid != identity.signing_kid:
        raise TransportError("Key provider signing_kid does not match local identity metadata")
    if keys.encryption_kid and keys.encryption_kid != identity.encryption_kid:
        raise TransportError("Key provider encryption_kid does not match local identity metadata")
    return AgentIdentity(
        agent_id=identity.agent_id,
        signing_private_key=keys.signing_private_key,
        signing_public_key=identity.signing_public_key,
        encryption_private_key=keys.encryption_private_key,
        encryption_public_key=identity.encryption_public_key,
        signing_kid=identity.signing_kid,
        encryption_kid=identity.encryption_kid,
    )


@dataclass
class ResolvedRecipient:
    recipient: str
    public_key: str
    identity_document: dict[str, Any]
    delivery_channel: str
    endpoint: str | None = None
    amqp_service: dict[str, Any] | None = None
    mqtt_service: dict[str, Any] | None = None


def _reason_for_capability_mismatch(reason: str | None) -> FailReason:
    reason_lower = (reason or "").lower()
    if "protocol" in reason_lower:
        return FailReason.UNSUPPORTED_VERSION
    if "crypto" in reason_lower:
        return FailReason.UNSUPPORTED_CRYPTO_SUITE
    if "profile" in reason_lower:
        return FailReason.UNSUPPORTED_PROFILE
    return FailReason.POLICY_REJECTED


def _validate_http_config(
    *,
    url: str,
    policy: HttpSecurityPolicy,
    context: str,
) -> None:
    try:
        warning_messages = enforce_http_security(url, policy=policy, context=context)
    except HttpSecurityError as exc:
        raise TransportError(str(exc)) from exc
    for warning_message in warning_messages:
        warnings.warn(warning_message, RuntimeWarning, stacklevel=3)


def _delivery_state_from_response(
    *,
    status_code: int,
    response_class: MessageClass | None,
    reason_code: str | None,
) -> DeliveryState:
    if 200 <= status_code < 300:
        if response_class is MessageClass.FAIL:
            if reason_code == FailReason.EXPIRED_MESSAGE.value:
                return DeliveryState.EXPIRED
            if reason_code == FailReason.POLICY_REJECTED.value:
                return DeliveryState.DECLINED
            return DeliveryState.FAILED
        if response_class in {MessageClass.ACK, MessageClass.CAPABILITIES}:
            return DeliveryState.ACKNOWLEDGED
        return DeliveryState.DELIVERED
    if status_code == 410:
        return DeliveryState.EXPIRED
    if status_code in {401, 403, 409, 422}:
        return DeliveryState.DECLINED
    return DeliveryState.FAILED


def _base_url_from_endpoint(endpoint: Any) -> str | None:
    if not isinstance(endpoint, str) or not endpoint.strip():
        return None
    parsed = urlsplit(endpoint.strip())
    if parsed.scheme not in {"http", "https"} or not parsed.netloc:
        return None
    return f"{parsed.scheme}://{parsed.netloc}"


class Agent:
    def __init__(
        self,
        *,
        identity: AgentIdentity,
        identity_document: dict[str, Any],
        discovery: DiscoveryClient,
        relay_client: RelayClient,
        capabilities: AgentCapabilities,
        storage_dir: Path,
        trust_profile: str,
        amqp_transport: AMQPTransport | None,
        mqtt_transport: MQTTTransport | None,
        key_provider_info: dict[str, Any] | None,
    ) -> None:
        self.identity = identity
        self.identity_document = identity_document
        self.discovery = discovery
        self.relay_client = relay_client
        self.capabilities = capabilities
        self.storage_dir = storage_dir
        self.trust_profile = trust_profile
        self.amqp_transport = amqp_transport
        self.mqtt_transport = mqtt_transport
        self.key_provider_info = dict(key_provider_info or {})
        self.delivery_states: dict[str, dict[str, str]] = {}
        self._processed_message_ids: set[str] = set()

    @property
    def agent_id(self) -> str:
        return self.identity.agent_id

    def build_well_known_document(
        self,
        *,
        base_url: str | None = None,
        identity_document_url: str | None = None,
    ) -> dict[str, Any]:
        resolved_base_url = base_url or _base_url_from_endpoint(
            self.identity_document.get("service", {}).get("direct_endpoint"),
        )
        if not isinstance(resolved_base_url, str) or not resolved_base_url.strip():
            raise TransportError(
                "Unable to build /.well-known/acp metadata without base_url or direct_endpoint",
            )
        return build_well_known_metadata(
            identity_document=self.identity_document,
            base_url=resolved_base_url,
            identity_document_url=identity_document_url,
        )

    @classmethod
    def create(
        cls,
        agent_id: str,
        *,
        storage_dir: str | Path = ".acp-data",
        endpoint: str | None = None,
        relay_url: str = "https://localhost:8080",
        relay_hints: list[str] | None = None,
        enterprise_directory_hints: list[str] | None = None,
        discovery_scheme: str = "https",
        trust_profile: str = "self_asserted",
        capabilities: AgentCapabilities | None = None,
        amqp_broker_url: str | None = None,
        amqp_exchange: str = DEFAULT_AMQP_EXCHANGE,
        amqp_exchange_type: str = "direct",
        mqtt_broker_url: str | None = None,
        mqtt_qos: int = DEFAULT_MQTT_QOS,
        mqtt_topic_prefix: str = DEFAULT_MQTT_TOPIC_PREFIX,
        allow_insecure_http: bool = False,
        allow_insecure_tls: bool = False,
        ca_file: str | None = None,
        mtls_enabled: bool = False,
        cert_file: str | None = None,
        key_file: str | None = None,
        key_provider: KeyProvider | None = None,
    ) -> "Agent":
        return cls.load_or_create(
            agent_id,
            storage_dir=storage_dir,
            endpoint=endpoint,
            relay_url=relay_url,
            relay_hints=relay_hints,
            enterprise_directory_hints=enterprise_directory_hints,
            discovery_scheme=discovery_scheme,
            trust_profile=trust_profile,
            capabilities=capabilities,
            amqp_broker_url=amqp_broker_url,
            amqp_exchange=amqp_exchange,
            amqp_exchange_type=amqp_exchange_type,
            mqtt_broker_url=mqtt_broker_url,
            mqtt_qos=mqtt_qos,
            mqtt_topic_prefix=mqtt_topic_prefix,
            allow_insecure_http=allow_insecure_http,
            allow_insecure_tls=allow_insecure_tls,
            ca_file=ca_file,
            mtls_enabled=mtls_enabled,
            cert_file=cert_file,
            key_file=key_file,
            key_provider=key_provider,
        )

    @classmethod
    def load_or_create(
        cls,
        agent_id: str,
        *,
        storage_dir: str | Path = ".acp-data",
        endpoint: str | None = None,
        relay_url: str = "https://localhost:8080",
        relay_hints: list[str] | None = None,
        enterprise_directory_hints: list[str] | None = None,
        discovery_scheme: str = "https",
        trust_profile: str = "self_asserted",
        capabilities: AgentCapabilities | None = None,
        amqp_broker_url: str | None = None,
        amqp_exchange: str = DEFAULT_AMQP_EXCHANGE,
        amqp_exchange_type: str = "direct",
        mqtt_broker_url: str | None = None,
        mqtt_qos: int = DEFAULT_MQTT_QOS,
        mqtt_topic_prefix: str = DEFAULT_MQTT_TOPIC_PREFIX,
        allow_insecure_http: bool = False,
        allow_insecure_tls: bool = False,
        ca_file: str | None = None,
        mtls_enabled: bool = False,
        cert_file: str | None = None,
        key_file: str | None = None,
        key_provider: KeyProvider | None = None,
    ) -> "Agent":
        parse_agent_id(agent_id)
        storage = Path(storage_dir)
        storage.mkdir(parents=True, exist_ok=True)
        provider = key_provider or LocalKeyProvider(
            storage_dir=storage,
            cert_file=cert_file,
            key_file=key_file,
            ca_file=ca_file,
        )
        try:
            key_provider_info = provider.describe()
        except Exception:  # noqa: BLE001
            key_provider_info = {"provider": "unknown"}

        provider_tls_material = None
        provider_ca_bundle: str | None = None
        try:
            provider_tls_material = provider.load_tls_material(agent_id)
        except KeyProviderError as exc:
            if mtls_enabled:
                raise TransportError(f"Unable to load TLS material from key provider: {exc}") from exc
        try:
            provider_ca_bundle = provider.load_ca_bundle(agent_id)
        except KeyProviderError:
            provider_ca_bundle = None

        effective_ca_file = (
            ca_file
            or (
                provider_tls_material.ca_file
                if provider_tls_material is not None and provider_tls_material.ca_file
                else None
            )
            or provider_ca_bundle
        )
        effective_cert_file = (
            cert_file
            or (
                provider_tls_material.cert_file
                if provider_tls_material is not None and provider_tls_material.cert_file
                else None
            )
        )
        effective_key_file = (
            key_file
            or (
                provider_tls_material.key_file
                if provider_tls_material is not None and provider_tls_material.key_file
                else None
            )
        )

        http_policy = HttpSecurityPolicy(
            allow_insecure_http=allow_insecure_http,
            allow_insecure_tls=allow_insecure_tls,
            ca_file=effective_ca_file,
            mtls_enabled=mtls_enabled,
            cert_file=effective_cert_file,
            key_file=effective_key_file,
        )
        try:
            warning_messages = validate_http_security_policy(
                http_policy,
                context="Agent configuration",
            )
        except HttpSecurityError as exc:
            raise TransportError(str(exc)) from exc
        for warning_message in warning_messages:
            warnings.warn(warning_message, RuntimeWarning, stacklevel=3)
        if endpoint:
            _validate_http_config(
                url=endpoint,
                policy=http_policy,
                context="Agent direct endpoint configuration",
            )
        if relay_url:
            _validate_http_config(
                url=relay_url,
                policy=http_policy,
                context="Agent relay URL configuration",
            )
        for relay_hint in relay_hints or []:
            _validate_http_config(
                url=str(relay_hint),
                policy=http_policy,
                context="Agent relay hint configuration",
            )
        for directory_hint in enterprise_directory_hints or []:
            _validate_http_config(
                url=str(directory_hint),
                policy=http_policy,
                context="Agent enterprise directory hint configuration",
            )

        local_amqp_service = (
            build_amqp_service_hint(
                agent_id=agent_id,
                broker_url=amqp_broker_url,
                exchange=amqp_exchange,
            )
            if amqp_broker_url
            else None
        )
        local_mqtt_service = (
            build_mqtt_service_hint(
                agent_id=agent_id,
                broker_url=mqtt_broker_url,
                qos=mqtt_qos,
                topic_prefix=mqtt_topic_prefix,
            )
            if mqtt_broker_url
            else None
        )
        http_security_profile = "mtls" if mtls_enabled else None

        provider_identity_keys: IdentityKeyMaterial | None = None
        provider_identity_error: KeyProviderError | None = None
        try:
            provider_identity_keys = provider.load_identity_keys(agent_id)
        except KeyProviderError as exc:
            provider_identity_error = exc

        existing = read_identity(storage, agent_id)
        if existing is None:
            if provider_identity_keys is not None:
                identity = _identity_from_provider(agent_id, provider_identity_keys)
            elif key_provider is not None and not isinstance(provider, LocalKeyProvider):
                raise TransportError(
                    f"Unable to load identity keys from key provider: {provider_identity_error}",
                )
            else:
                identity = AgentIdentity.create(agent_id)
            capabilities_obj = capabilities or AgentCapabilities(agent_id=agent_id)
            identity_document = identity.build_identity_document(
                direct_endpoint=endpoint,
                relay_hints=relay_hints,
                http_security_profile=http_security_profile,
                relay_security_profile=http_security_profile,
                amqp_service=local_amqp_service,
                mqtt_service=local_mqtt_service,
                trust_profile=trust_profile,
                capabilities=capabilities_obj.to_dict(),
            )
            write_identity(storage, identity, identity_document)
        else:
            identity, identity_document = existing
            if provider_identity_keys is not None:
                identity = _apply_provider_keys(identity, provider_identity_keys)
            elif key_provider is not None and not isinstance(provider, LocalKeyProvider):
                raise TransportError(
                    f"Unable to load identity keys from key provider: {provider_identity_error}",
                )
            if not verify_identity_document(identity_document):
                capabilities_obj = capabilities or AgentCapabilities(agent_id=agent_id)
                identity_document = identity.build_identity_document(
                    direct_endpoint=endpoint,
                    relay_hints=relay_hints,
                    http_security_profile=http_security_profile,
                    relay_security_profile=http_security_profile,
                    amqp_service=local_amqp_service,
                    mqtt_service=local_mqtt_service,
                    trust_profile=trust_profile,
                    capabilities=capabilities_obj.to_dict(),
                )
                write_identity(storage, identity, identity_document)
            else:
                capabilities_obj = capabilities or AgentCapabilities.from_dict(
                    identity_document.get("capabilities"),
                    fallback_agent_id=agent_id,
                )
                if (
                    endpoint is not None
                    or relay_hints is not None
                    or capabilities is not None
                    or local_amqp_service is not None
                    or local_mqtt_service is not None
                ):
                    identity_document = identity.build_identity_document(
                        direct_endpoint=endpoint
                        if endpoint is not None
                        else identity_document.get("service", {}).get("direct_endpoint"),
                        relay_hints=relay_hints
                        if relay_hints is not None
                        else identity_document.get("service", {}).get("relay_hints", []),
                        http_security_profile=http_security_profile,
                        relay_security_profile=http_security_profile,
                        amqp_service=local_amqp_service
                        if local_amqp_service is not None
                        else identity_document.get("service", {}).get("amqp"),
                        mqtt_service=local_mqtt_service
                        if local_mqtt_service is not None
                        else identity_document.get("service", {}).get("mqtt"),
                        trust_profile=trust_profile,
                        capabilities=capabilities_obj.to_dict(),
                    )
                    write_identity(storage, identity, identity_document)

        effective_hints = relay_hints if relay_hints is not None else identity_document.get(
            "service",
            {},
        ).get(
            "relay_hints",
            [],
        )
        if relay_url and relay_url not in effective_hints:
            effective_hints = [*effective_hints, relay_url]

        discovery = DiscoveryClient(
            cache_path=storage / "discovery_cache.json",
            default_scheme=discovery_scheme,
            relay_hints=effective_hints,
            enterprise_directory_hints=enterprise_directory_hints,
            allow_insecure_http=allow_insecure_http,
            allow_insecure_tls=allow_insecure_tls,
            ca_file=effective_ca_file,
            mtls_enabled=mtls_enabled,
            cert_file=effective_cert_file,
            key_file=effective_key_file,
        )
        discovery.seed(identity_document)

        amqp_transport: AMQPTransport | None = None
        if amqp_broker_url:
            amqp_transport = AMQPTransport(
                broker_url=amqp_broker_url,
                exchange=amqp_exchange,
                exchange_type=amqp_exchange_type,
            )

        mqtt_transport: MQTTTransport | None = None
        if mqtt_broker_url:
            mqtt_transport = MQTTTransport(
                broker_url=mqtt_broker_url,
                qos=mqtt_qos,
                topic_prefix=mqtt_topic_prefix,
            )

        return cls(
            identity=identity,
            identity_document=identity_document,
            discovery=discovery,
            relay_client=RelayClient(
                relay_url,
                allow_insecure_http=allow_insecure_http,
                allow_insecure_tls=allow_insecure_tls,
                ca_file=effective_ca_file,
                mtls_enabled=mtls_enabled,
                cert_file=effective_cert_file,
                key_file=effective_key_file,
            ),
            capabilities=capabilities_obj,
            storage_dir=storage,
            trust_profile=trust_profile,
            amqp_transport=amqp_transport,
            mqtt_transport=mqtt_transport,
            key_provider_info=key_provider_info,
        )

    def _shared_transports(self, remote_capabilities: AgentCapabilities) -> list[str]:
        remote_transports = {item.lower() for item in remote_capabilities.transports}
        shared: list[str] = []
        for transport in self.capabilities.transports:
            normalized = transport.lower()
            if normalized in remote_transports:
                shared.append(normalized)
        return shared

    def _choose_delivery_channel(
        self,
        *,
        remote_capabilities: AgentCapabilities,
        identity_doc: dict[str, Any],
        delivery_mode: str,
    ) -> tuple[str | None, str | None, dict[str, Any] | None, dict[str, Any] | None]:
        shared_transports = self._shared_transports(remote_capabilities)
        direct_endpoint = identity_doc.get("service", {}).get("direct_endpoint")
        has_direct_endpoint = isinstance(direct_endpoint, str) and bool(direct_endpoint.strip())
        amqp_service = identity_doc.get("service", {}).get("amqp")
        amqp_service_valid = (
            isinstance(amqp_service, dict)
            and isinstance(amqp_service.get("broker_url"), str)
            and bool(str(amqp_service.get("broker_url")).strip())
        )
        amqp_available = "amqp" in shared_transports and amqp_service_valid
        mqtt_service = identity_doc.get("service", {}).get("mqtt")
        mqtt_service_valid = (
            isinstance(mqtt_service, dict)
            and isinstance(mqtt_service.get("broker_url"), str)
            and bool(str(mqtt_service.get("broker_url")).strip())
            and isinstance(mqtt_service.get("topic"), str)
            and bool(str(mqtt_service.get("topic")).strip())
        )
        mqtt_available = "mqtt" in shared_transports and mqtt_service_valid

        def _direct_available() -> bool:
            if not has_direct_endpoint:
                return False
            return any(transport in {"https", "http", "direct"} for transport in shared_transports)

        if delivery_mode == "direct":
            if _direct_available():
                return "direct", str(direct_endpoint), None, None
            return None, "No compatible direct transport and endpoint available", None, None

        if delivery_mode == "relay":
            if "relay" in shared_transports:
                return "relay", None, None, None
            return None, "No compatible relay transport available", None, None

        if delivery_mode == "amqp":
            if amqp_available:
                return "amqp", None, dict(amqp_service), None
            return None, "No compatible AMQP transport available", None, None

        if delivery_mode == "mqtt":
            if mqtt_available:
                return "mqtt", None, None, dict(mqtt_service)
            return None, "No compatible MQTT transport available", None, None

        for transport in shared_transports:
            if transport in {"https", "http", "direct"} and _direct_available():
                return "direct", str(direct_endpoint), None, None
            if transport == "relay":
                return "relay", None, None, None
            if transport == "amqp" and amqp_available:
                return "amqp", None, dict(amqp_service), None
            if transport == "mqtt" and mqtt_available:
                return "mqtt", None, None, dict(mqtt_service)

        if has_direct_endpoint:
            return None, "No compatible transport implementation available for this recipient", None, None
        if amqp_service_valid:
            return (
                None,
                "AMQP transport is advertised but not compatible with sender capabilities",
                None,
                None,
            )
        if mqtt_service_valid:
            return (
                None,
                "MQTT transport is advertised but not compatible with sender capabilities",
                None,
                None,
            )
        return (
            None,
            "Recipient identity document is missing direct_endpoint/amqp/mqtt and no relay fallback is compatible",
            None,
            None,
        )

    def _build_message(
        self,
        *,
        recipients: list[str],
        payload: dict[str, Any],
        recipient_public_keys: dict[str, str],
        message_class: MessageClass,
        context_id: str,
        operation_id: str,
        expires_in_seconds: int,
        correlation_id: str | None,
        in_reply_to: str | None,
    ) -> ACPMessage:
        envelope = Envelope.build(
            sender=self.agent_id,
            recipients=recipients,
            message_class=message_class,
            context_id=context_id,
            expires_in_seconds=expires_in_seconds,
            operation_id=operation_id,
            correlation_id=correlation_id,
            in_reply_to=in_reply_to,
        )
        protected = encrypt_for_recipients(payload, envelope, recipient_public_keys)
        protected = sign_protected_payload(
            envelope,
            protected,
            self.identity.signing_private_key,
            self.identity.signing_kid,
        )
        return ACPMessage(
            envelope=envelope,
            protected=protected,
            sender_identity_document=self.identity_document,
        )

    def _resolve_sender_identity_document(
        self,
        *,
        raw_message: dict[str, Any],
        sender_id: str,
    ) -> dict[str, Any]:
        embedded_identity_document = raw_message.get("sender_identity_document")
        if (
            isinstance(embedded_identity_document, dict)
            and embedded_identity_document.get("agent_id") == sender_id
            and verify_identity_document(embedded_identity_document)
        ):
            return embedded_identity_document
        return self.discovery.resolve(sender_id)

    def _resolve_recipients(
        self,
        recipients: list[str],
        delivery_mode: str,
    ) -> tuple[list[ResolvedRecipient], list[DeliveryOutcome]]:
        resolved: list[ResolvedRecipient] = []
        outcomes: list[DeliveryOutcome] = []

        for recipient in recipients:
            try:
                identity_doc = self.discovery.resolve(recipient)
            except DiscoveryError as exc:
                outcomes.append(
                    DeliveryOutcome(
                        recipient=recipient,
                        state=DeliveryState.FAILED,
                        reason_code=FailReason.POLICY_REJECTED.value,
                        detail=str(exc),
                    ),
                )
                continue

            remote_capabilities = AgentCapabilities.from_dict(
                identity_doc.get("capabilities"),
                fallback_agent_id=recipient,
            )
            capability_match = choose_compatible(self.capabilities, remote_capabilities)
            if not capability_match.is_compatible:
                fail_reason = _reason_for_capability_mismatch(capability_match.reason)
                outcomes.append(
                    DeliveryOutcome(
                        recipient=recipient,
                        state=DeliveryState.FAILED,
                        reason_code=fail_reason.value,
                        detail=capability_match.reason,
                    ),
                )
                continue

            delivery_channel, endpoint, amqp_service, mqtt_service = self._choose_delivery_channel(
                remote_capabilities=remote_capabilities,
                identity_doc=identity_doc,
                delivery_mode=delivery_mode,
            )
            if delivery_channel is None:
                outcomes.append(
                    DeliveryOutcome(
                        recipient=recipient,
                        state=DeliveryState.FAILED,
                        reason_code=FailReason.POLICY_REJECTED.value,
                        detail=endpoint,
                    ),
                )
                continue

            recipient_public_key = (
                identity_doc.get("keys", {}).get("encryption", {}).get("public_key")
            )
            if not isinstance(recipient_public_key, str):
                outcomes.append(
                    DeliveryOutcome(
                        recipient=recipient,
                        state=DeliveryState.FAILED,
                        reason_code=FailReason.POLICY_REJECTED.value,
                        detail="Recipient identity document missing encryption public key",
                    ),
                )
                continue

            resolved.append(
                ResolvedRecipient(
                    recipient=recipient,
                    public_key=recipient_public_key,
                    identity_document=identity_doc,
                    delivery_channel=delivery_channel,
                    endpoint=endpoint,
                    amqp_service=amqp_service,
                    mqtt_service=mqtt_service,
                ),
            )

        return resolved, outcomes

    def _outcome_from_http_response(
        self,
        *,
        recipient: str,
        status_code: int,
        body: dict[str, Any] | None,
    ) -> DeliveryOutcome:
        response_message = body.get("response_message") if isinstance(body, dict) else None
        response_class_raw = (
            response_message.get("envelope", {}).get("message_class")
            if isinstance(response_message, dict)
            else None
        )
        response_class: MessageClass | None = None
        if isinstance(response_class_raw, str):
            try:
                response_class = MessageClass(response_class_raw)
            except ValueError:
                response_class = None

        reason_code = body.get("reason_code") if isinstance(body, dict) else None
        detail = body.get("detail") if isinstance(body, dict) else None
        if not isinstance(reason_code, str):
            reason_code = None
        if not isinstance(detail, str):
            detail = None
        if detail is None and status_code >= 400:
            detail = f"Recipient HTTP {status_code}"

        return DeliveryOutcome(
            recipient=recipient,
            state=_delivery_state_from_response(
                status_code=status_code,
                response_class=response_class,
                reason_code=reason_code,
            ),
            status_code=status_code,
            response_class=response_class,
            reason_code=reason_code,
            detail=detail,
            response_message=response_message if isinstance(response_message, dict) else None,
        )

    def _deliver_direct(
        self,
        *,
        message: ACPMessage,
        targets: list[ResolvedRecipient],
    ) -> list[DeliveryOutcome]:
        outcomes: list[DeliveryOutcome] = []
        for target in targets:
            if target.endpoint is None:
                outcomes.append(
                    DeliveryOutcome(
                        recipient=target.recipient,
                        state=DeliveryState.FAILED,
                        reason_code=FailReason.POLICY_REJECTED.value,
                        detail="Missing direct endpoint for direct delivery",
                    ),
                )
                continue
            try:
                response = self.relay_client.transport.post_json(target.endpoint, message.to_dict())
            except TransportError as exc:
                outcomes.append(
                    DeliveryOutcome(
                        recipient=target.recipient,
                        state=DeliveryState.FAILED,
                        reason_code=FailReason.POLICY_REJECTED.value,
                        detail=f"Direct transport failure: {exc}",
                    ),
                )
                continue
            try:
                body = response.json()
            except ValueError:
                body = None
            outcomes.append(
                self._outcome_from_http_response(
                    recipient=target.recipient,
                    status_code=response.status_code,
                    body=body if isinstance(body, dict) else None,
                ),
            )
        return outcomes

    def _deliver_via_relay(
        self,
        *,
        message: ACPMessage,
        targets: list[ResolvedRecipient],
    ) -> list[DeliveryOutcome]:
        expected_recipients = {target.recipient for target in targets}
        outcomes: list[DeliveryOutcome] = []
        try:
            relay_response = self.relay_client.send_message(message)
        except TransportError as exc:
            for recipient in expected_recipients:
                outcomes.append(
                    DeliveryOutcome(
                        recipient=recipient,
                        state=DeliveryState.FAILED,
                        reason_code=FailReason.POLICY_REJECTED.value,
                        detail=f"Relay transport failure: {exc}",
                    ),
                )
            return outcomes

        delivered: set[str] = set()
        for raw_outcome in relay_response.get("outcomes", []):
            try:
                outcome = DeliveryOutcome.from_dict(raw_outcome)
            except Exception:  # noqa: BLE001
                continue
            delivered.add(outcome.recipient)
            outcomes.append(outcome)

        for recipient in expected_recipients - delivered:
            outcomes.append(
                DeliveryOutcome(
                    recipient=recipient,
                    state=DeliveryState.FAILED,
                    reason_code=FailReason.POLICY_REJECTED.value,
                    detail="Relay did not return an outcome for recipient",
                ),
            )
        return outcomes

    def _deliver_via_amqp(
        self,
        *,
        payload: dict[str, Any],
        message_class: MessageClass,
        context_id: str,
        operation_id: str,
        expires_in_seconds: int,
        correlation_id: str | None,
        in_reply_to: str | None,
        targets: list[ResolvedRecipient],
    ) -> tuple[list[DeliveryOutcome], list[str]]:
        outcomes: list[DeliveryOutcome] = []
        message_ids: list[str] = []

        if self.amqp_transport is None:
            for target in targets:
                outcomes.append(
                    DeliveryOutcome(
                        recipient=target.recipient,
                        state=DeliveryState.FAILED,
                        reason_code=FailReason.POLICY_REJECTED.value,
                        detail=(
                            "AMQP delivery selected but sender is not configured with AMQP broker settings"
                        ),
                    ),
                )
            return outcomes, message_ids

        for target in targets:
            outbound_message = self._build_message(
                recipients=[target.recipient],
                payload=payload,
                recipient_public_keys={target.recipient: target.public_key},
                message_class=message_class,
                context_id=context_id,
                operation_id=operation_id,
                expires_in_seconds=expires_in_seconds,
                correlation_id=correlation_id,
                in_reply_to=in_reply_to,
            )
            message_ids.append(outbound_message.envelope.message_id)
            try:
                self.amqp_transport.publish(
                    message=outbound_message.to_dict(),
                    recipient_agent_id=target.recipient,
                    amqp_service=target.amqp_service,
                )
                outcomes.append(
                    DeliveryOutcome(
                        recipient=target.recipient,
                        state=DeliveryState.DELIVERED,
                    ),
                )
            except AMQPTransportError as exc:
                outcomes.append(
                    DeliveryOutcome(
                        recipient=target.recipient,
                        state=DeliveryState.FAILED,
                        reason_code=FailReason.POLICY_REJECTED.value,
                        detail=f"AMQP transport failure: {exc}",
                    ),
                )
        return outcomes, message_ids

    def _deliver_via_mqtt(
        self,
        *,
        payload: dict[str, Any],
        message_class: MessageClass,
        context_id: str,
        operation_id: str,
        expires_in_seconds: int,
        correlation_id: str | None,
        in_reply_to: str | None,
        targets: list[ResolvedRecipient],
    ) -> tuple[list[DeliveryOutcome], list[str]]:
        outcomes: list[DeliveryOutcome] = []
        message_ids: list[str] = []

        if self.mqtt_transport is None:
            for target in targets:
                outcomes.append(
                    DeliveryOutcome(
                        recipient=target.recipient,
                        state=DeliveryState.FAILED,
                        reason_code=FailReason.POLICY_REJECTED.value,
                        detail=(
                            "MQTT delivery selected but sender is not configured with MQTT broker settings"
                        ),
                    ),
                )
            return outcomes, message_ids

        for target in targets:
            outbound_message = self._build_message(
                recipients=[target.recipient],
                payload=payload,
                recipient_public_keys={target.recipient: target.public_key},
                message_class=message_class,
                context_id=context_id,
                operation_id=operation_id,
                expires_in_seconds=expires_in_seconds,
                correlation_id=correlation_id,
                in_reply_to=in_reply_to,
            )
            message_ids.append(outbound_message.envelope.message_id)
            try:
                self.mqtt_transport.publish(
                    message=outbound_message.to_dict(),
                    recipient_agent_id=target.recipient,
                    mqtt_service=target.mqtt_service,
                )
                outcomes.append(
                    DeliveryOutcome(
                        recipient=target.recipient,
                        state=DeliveryState.DELIVERED,
                    ),
                )
            except MQTTTransportError as exc:
                outcomes.append(
                    DeliveryOutcome(
                        recipient=target.recipient,
                        state=DeliveryState.FAILED,
                        reason_code=FailReason.POLICY_REJECTED.value,
                        detail=f"MQTT transport failure: {exc}",
                    ),
                )
        return outcomes, message_ids

    def _sync_delivery_states(self, operation_id: str, outcomes: list[DeliveryOutcome]) -> None:
        self.delivery_states[operation_id] = {
            outcome.recipient: outcome.state.value for outcome in outcomes
        }

    def send(
        self,
        *,
        recipients: list[str],
        payload: dict[str, Any],
        context: str | None = None,
        message_class: MessageClass = MessageClass.SEND,
        expires_in_seconds: int = 300,
        correlation_id: str | None = None,
        in_reply_to: str | None = None,
        delivery_mode: str = "auto",
    ) -> SendResult:
        if not recipients:
            raise ValueError("send() requires at least one recipient")
        if delivery_mode not in {"auto", "direct", "relay", "amqp", "mqtt"}:
            raise ValueError("delivery_mode must be one of: auto, direct, relay, amqp, mqtt")

        operation_id = str(uuid.uuid4())
        context_id = context or operation_id

        resolved_recipients, preflight_outcomes = self._resolve_recipients(
            recipients,
            delivery_mode=delivery_mode,
        )
        if not resolved_recipients:
            result = SendResult(
                operation_id=operation_id,
                message_id=str(uuid.uuid4()),
                outcomes=preflight_outcomes,
            )
            self._sync_delivery_states(operation_id, preflight_outcomes)
            return result

        outcomes = [*preflight_outcomes]
        outbound_message_ids: list[str] = []

        direct_targets = [
            target for target in resolved_recipients if target.delivery_channel == "direct"
        ]
        relay_targets = [
            target for target in resolved_recipients if target.delivery_channel == "relay"
        ]
        amqp_targets = [
            target for target in resolved_recipients if target.delivery_channel == "amqp"
        ]
        mqtt_targets = [
            target for target in resolved_recipients if target.delivery_channel == "mqtt"
        ]

        if direct_targets:
            direct_message = self._build_message(
                recipients=[target.recipient for target in direct_targets],
                payload=payload,
                recipient_public_keys={
                    target.recipient: target.public_key for target in direct_targets
                },
                message_class=message_class,
                context_id=context_id,
                operation_id=operation_id,
                expires_in_seconds=expires_in_seconds,
                correlation_id=correlation_id,
                in_reply_to=in_reply_to,
            )
            outbound_message_ids.append(direct_message.envelope.message_id)
            outcomes.extend(
                self._deliver_direct(
                    message=direct_message,
                    targets=direct_targets,
                ),
            )

        if relay_targets:
            relay_message = self._build_message(
                recipients=[target.recipient for target in relay_targets],
                payload=payload,
                recipient_public_keys={
                    target.recipient: target.public_key for target in relay_targets
                },
                message_class=message_class,
                context_id=context_id,
                operation_id=operation_id,
                expires_in_seconds=expires_in_seconds,
                correlation_id=correlation_id,
                in_reply_to=in_reply_to,
            )
            outbound_message_ids.append(relay_message.envelope.message_id)
            outcomes.extend(
                self._deliver_via_relay(
                    message=relay_message,
                    targets=relay_targets,
                ),
            )

        if amqp_targets:
            amqp_outcomes, amqp_message_ids = self._deliver_via_amqp(
                payload=payload,
                message_class=message_class,
                context_id=context_id,
                operation_id=operation_id,
                expires_in_seconds=expires_in_seconds,
                correlation_id=correlation_id,
                in_reply_to=in_reply_to,
                targets=amqp_targets,
            )
            outbound_message_ids.extend(amqp_message_ids)
            outcomes.extend(amqp_outcomes)

        if mqtt_targets:
            mqtt_outcomes, mqtt_message_ids = self._deliver_via_mqtt(
                payload=payload,
                message_class=message_class,
                context_id=context_id,
                operation_id=operation_id,
                expires_in_seconds=expires_in_seconds,
                correlation_id=correlation_id,
                in_reply_to=in_reply_to,
                targets=mqtt_targets,
            )
            outbound_message_ids.extend(mqtt_message_ids)
            outcomes.extend(mqtt_outcomes)

        if not outbound_message_ids:
            outbound_message_ids.append(str(uuid.uuid4()))

        result = SendResult(
            operation_id=operation_id,
            message_id=outbound_message_ids[0],
            outcomes=outcomes,
            message_ids=outbound_message_ids,
        )
        self._sync_delivery_states(operation_id, outcomes)
        return result

    def send_compensate(
        self,
        *,
        recipients: list[str],
        original_operation_id: str,
        reason: str,
        actions: list[dict[str, Any]] | None = None,
        context: str | None = None,
        delivery_mode: str = "auto",
    ) -> SendResult:
        instruction = CompensateInstruction(
            operation_id=original_operation_id,
            reason=reason,
            actions=actions or [],
        )
        return self.send(
            recipients=recipients,
            payload={"compensation": instruction.to_dict()},
            context=context or f"compensate:{original_operation_id}",
            message_class=MessageClass.COMPENSATE,
            correlation_id=original_operation_id,
            delivery_mode=delivery_mode,
        )

    def _create_response_message(
        self,
        *,
        sender_identity_document: dict[str, Any],
        request_envelope: Envelope,
        response_message_class: MessageClass,
        response_payload: dict[str, Any],
    ) -> ACPMessage:
        sender_id = request_envelope.sender
        sender_encryption_public_key = (
            sender_identity_document.get("keys", {}).get("encryption", {}).get("public_key")
        )
        if not isinstance(sender_encryption_public_key, str):
            raise ProcessingError(
                reason_code=FailReason.POLICY_REJECTED,
                detail="Sender identity document missing encryption key",
            )

        return self._build_message(
            recipients=[sender_id],
            payload=response_payload,
            recipient_public_keys={sender_id: sender_encryption_public_key},
            message_class=response_message_class,
            context_id=request_envelope.context_id,
            operation_id=request_envelope.operation_id,
            expires_in_seconds=300,
            correlation_id=request_envelope.correlation_id or request_envelope.operation_id,
            in_reply_to=request_envelope.message_id,
        )

    def decrypt_message_for_self(self, raw_message: dict[str, Any]) -> tuple[ACPMessage, dict[str, Any]]:
        message = ACPMessage.from_dict(raw_message)
        if message.envelope.acp_version != ACP_VERSION:
            raise ProcessingError(
                reason_code=FailReason.UNSUPPORTED_VERSION,
                detail=f"Unsupported ACP version: {message.envelope.acp_version}",
            )
        if message.envelope.crypto_suite != DEFAULT_CRYPTO_SUITE:
            raise ProcessingError(
                reason_code=FailReason.UNSUPPORTED_CRYPTO_SUITE,
                detail=f"Unsupported crypto suite: {message.envelope.crypto_suite}",
            )
        if self.agent_id not in message.envelope.recipients:
            raise ProcessingError(
                reason_code=FailReason.POLICY_REJECTED,
                detail="Message is not addressed to this agent",
            )
        if message.envelope.is_expired():
            raise ProcessingError(
                reason_code=FailReason.EXPIRED_MESSAGE,
                detail=f"Message {message.envelope.message_id} is expired",
            )

        sender_doc = self._resolve_sender_identity_document(
            raw_message=raw_message,
            sender_id=message.envelope.sender,
        )
        sender_signing_key = sender_doc.get("keys", {}).get("signing", {}).get("public_key")
        if not isinstance(sender_signing_key, str):
            raise ProcessingError(
                reason_code=FailReason.INVALID_SIGNATURE,
                detail="Sender signing public key is missing from identity document",
            )
        if not verify_protected_payload_signature(
            message.envelope,
            message.protected,
            sender_signing_key,
        ):
            raise ProcessingError(
                reason_code=FailReason.INVALID_SIGNATURE,
                detail="Message signature verification failed",
            )

        try:
            payload = decrypt_for_recipient(
                message.envelope,
                message.protected,
                self.agent_id,
                self.identity.encryption_private_key,
            )
        except CryptoError as exc:
            raise ProcessingError(
                reason_code=FailReason.POLICY_REJECTED,
                detail=str(exc),
            ) from exc
        return message, payload

    def handle_incoming(
        self,
        raw_message: dict[str, Any],
        *,
        handler: IncomingHandler | None = None,
    ) -> dict[str, Any]:
        response_state = DeliveryState.FAILED
        decrypted_payload: dict[str, Any] | None = None
        reason_code: str | None = None
        detail: str | None = None
        response_message: ACPMessage | None = None

        try:
            request_message = ACPMessage.from_dict(raw_message)
        except Exception as exc:  # noqa: BLE001
            return {
                "state": DeliveryState.FAILED.value,
                "reason_code": FailReason.POLICY_REJECTED.value,
                "detail": f"Invalid ACP message structure: {exc}",
                "response_message": None,
            }

        sender_identity_document: dict[str, Any] | None = None
        try:
            if request_message.envelope.acp_version != ACP_VERSION:
                raise ProcessingError(
                    reason_code=FailReason.UNSUPPORTED_VERSION,
                    detail=f"Unsupported ACP version: {request_message.envelope.acp_version}",
                )
            if request_message.envelope.crypto_suite != DEFAULT_CRYPTO_SUITE:
                raise ProcessingError(
                    reason_code=FailReason.UNSUPPORTED_CRYPTO_SUITE,
                    detail=f"Unsupported crypto suite: {request_message.envelope.crypto_suite}",
                )
            if request_message.envelope.is_expired():
                raise ProcessingError(
                    reason_code=FailReason.EXPIRED_MESSAGE,
                    detail="Message is expired",
                )
            if self.agent_id not in request_message.envelope.recipients:
                raise ProcessingError(
                    reason_code=FailReason.POLICY_REJECTED,
                    detail=f"Recipient {self.agent_id} not in message recipients",
                )

            sender_identity_document = self._resolve_sender_identity_document(
                raw_message=raw_message,
                sender_id=request_message.envelope.sender,
            )
            sender_signing_key = (
                sender_identity_document.get("keys", {}).get("signing", {}).get("public_key")
            )
            if not isinstance(sender_signing_key, str):
                raise ProcessingError(
                    reason_code=FailReason.INVALID_SIGNATURE,
                    detail="Sender signing key missing from identity document",
                )

            if not verify_protected_payload_signature(
                request_message.envelope,
                request_message.protected,
                sender_signing_key,
            ):
                raise ProcessingError(
                    reason_code=FailReason.INVALID_SIGNATURE,
                    detail="Signature verification failed",
                )

            if request_message.envelope.message_id in self._processed_message_ids:
                response_state = DeliveryState.ACKNOWLEDGED
                if request_message.envelope.message_class not in {
                    MessageClass.ACK,
                    MessageClass.FAIL,
                }:
                    response_message = self._create_response_message(
                        sender_identity_document=sender_identity_document,
                        request_envelope=request_message.envelope,
                        response_message_class=MessageClass.ACK,
                        response_payload=build_ack_payload(
                            request_message.envelope.message_id,
                            status="duplicate",
                        ),
                    )
                return {
                    "state": response_state.value,
                    "reason_code": reason_code,
                    "detail": detail,
                    "decrypted_payload": decrypted_payload,
                    "response_message": response_message.to_dict() if response_message else None,
                }

            decrypted_payload = decrypt_for_recipient(
                request_message.envelope,
                request_message.protected,
                self.agent_id,
                self.identity.encryption_private_key,
            )

            if request_message.envelope.message_class is MessageClass.CAPABILITIES:
                response_state = DeliveryState.ACKNOWLEDGED
                response_payload = self.capabilities.to_dict()
                response_message = self._create_response_message(
                    sender_identity_document=sender_identity_document,
                    request_envelope=request_message.envelope,
                    response_message_class=MessageClass.CAPABILITIES,
                    response_payload=response_payload,
                )
            else:
                handler_payload: dict[str, Any] | None = None
                if handler is not None:
                    handler_payload = handler(decrypted_payload, request_message.envelope)
                response_state = DeliveryState.ACKNOWLEDGED
                if request_message.envelope.message_class not in {
                    MessageClass.ACK,
                    MessageClass.FAIL,
                }:
                    ack_payload = build_ack_payload(request_message.envelope.message_id)
                    if handler_payload:
                        ack_payload["handler"] = handler_payload
                    response_message = self._create_response_message(
                        sender_identity_document=sender_identity_document,
                        request_envelope=request_message.envelope,
                        response_message_class=MessageClass.ACK,
                        response_payload=ack_payload,
                    )
            self._processed_message_ids.add(request_message.envelope.message_id)
        except ProcessingError as exc:
            reason_code = exc.reason_code.value
            detail = exc.detail
        except DiscoveryError as exc:
            reason_code = FailReason.POLICY_REJECTED.value
            detail = f"Discovery failed for sender identity: {exc}"
        except CryptoError as exc:
            reason_code = FailReason.POLICY_REJECTED.value
            detail = str(exc)
        except Exception as exc:  # noqa: BLE001
            reason_code = FailReason.POLICY_REJECTED.value
            detail = str(exc)

        if response_state is DeliveryState.FAILED:
            if reason_code is None:
                reason_code = FailReason.POLICY_REJECTED.value
            if detail is None:
                detail = "Message processing failed"
            if sender_identity_document is not None:
                try:
                    fail_payload = build_fail_payload(reason_code=reason_code, detail=detail)
                    response_message = self._create_response_message(
                        sender_identity_document=sender_identity_document,
                        request_envelope=request_message.envelope,
                        response_message_class=MessageClass.FAIL,
                        response_payload=fail_payload,
                    )
                except Exception:  # noqa: BLE001
                    response_message = None

        return {
            "state": response_state.value,
            "reason_code": reason_code,
            "detail": detail,
            "decrypted_payload": decrypted_payload,
            "response_message": response_message.to_dict() if response_message else None,
        }

    def _publish_amqp_response_message(
        self,
        *,
        raw_message: dict[str, Any],
        response_message: dict[str, Any],
    ) -> None:
        if self.amqp_transport is None:
            raise AMQPTransportError("AMQP transport is not configured")
        envelope = raw_message.get("envelope")
        if not isinstance(envelope, dict):
            raise AMQPTransportError("Inbound message envelope is missing for AMQP response routing")
        sender_id = envelope.get("sender")
        if not isinstance(sender_id, str) or not sender_id.strip():
            raise AMQPTransportError("Inbound message sender is missing for AMQP response routing")

        sender_doc = self._resolve_sender_identity_document(
            raw_message=raw_message,
            sender_id=sender_id,
        )
        sender_service = sender_doc.get("service")
        sender_amqp_service = sender_service.get("amqp") if isinstance(sender_service, dict) else None
        if not isinstance(sender_amqp_service, dict):
            raise AMQPTransportError(
                f"Sender {sender_id} does not advertise service.amqp for AMQP response delivery",
            )

        self.amqp_transport.publish(
            message=response_message,
            recipient_agent_id=sender_id,
            amqp_service=sender_amqp_service,
        )

    def _publish_mqtt_response_message(
        self,
        *,
        raw_message: dict[str, Any],
        response_message: dict[str, Any],
    ) -> None:
        if self.mqtt_transport is None:
            raise MQTTTransportError("MQTT transport is not configured")
        envelope = raw_message.get("envelope")
        if not isinstance(envelope, dict):
            raise MQTTTransportError("Inbound message envelope is missing for MQTT response routing")
        sender_id = envelope.get("sender")
        if not isinstance(sender_id, str) or not sender_id.strip():
            raise MQTTTransportError("Inbound message sender is missing for MQTT response routing")

        sender_doc = self._resolve_sender_identity_document(
            raw_message=raw_message,
            sender_id=sender_id,
        )
        sender_service = sender_doc.get("service")
        sender_mqtt_service = sender_service.get("mqtt") if isinstance(sender_service, dict) else None
        if not isinstance(sender_mqtt_service, dict):
            raise MQTTTransportError(
                f"Sender {sender_id} does not advertise service.mqtt for MQTT response delivery",
            )

        self.mqtt_transport.publish(
            message=response_message,
            recipient_agent_id=sender_id,
            mqtt_service=sender_mqtt_service,
        )

    def consume_from_amqp(
        self,
        *,
        handler: IncomingHandler | None = None,
        max_messages: int | None = None,
    ) -> int:
        if self.amqp_transport is None:
            raise AMQPTransportError(
                "consume_from_amqp() requires an AMQP-configured agent (amqp_broker_url)",
            )

        service = self.identity_document.get("service", {})
        amqp_service = service.get("amqp") if isinstance(service, dict) else None
        if not isinstance(amqp_service, dict):
            raise AMQPTransportError("Identity document is missing service.amqp configuration")

        terminal_states = {
            DeliveryState.ACKNOWLEDGED.value,
            DeliveryState.FAILED.value,
            DeliveryState.DECLINED.value,
            DeliveryState.EXPIRED.value,
        }

        def _handle(raw_message: dict[str, Any]) -> bool:
            result = self.handle_incoming(raw_message, handler=handler)
            response_message = result.get("response_message")
            if isinstance(response_message, dict):
                try:
                    self._publish_amqp_response_message(
                        raw_message=raw_message,
                        response_message=response_message,
                    )
                except Exception:  # noqa: BLE001
                    return False
            return result.get("state") in terminal_states

        return self.amqp_transport.consume(
            agent_id=self.agent_id,
            handler=_handle,
            amqp_service=amqp_service,
            max_messages=max_messages,
        )

    def consume_from_mqtt(
        self,
        *,
        handler: IncomingHandler | None = None,
        max_messages: int | None = None,
    ) -> int:
        if self.mqtt_transport is None:
            raise MQTTTransportError(
                "consume_from_mqtt() requires an MQTT-configured agent (mqtt_broker_url)",
            )

        service = self.identity_document.get("service", {})
        mqtt_service = service.get("mqtt") if isinstance(service, dict) else None
        if not isinstance(mqtt_service, dict):
            raise MQTTTransportError("Identity document is missing service.mqtt configuration")

        terminal_states = {
            DeliveryState.ACKNOWLEDGED.value,
            DeliveryState.FAILED.value,
            DeliveryState.DECLINED.value,
            DeliveryState.EXPIRED.value,
        }

        def _handle(raw_message: dict[str, Any]) -> bool:
            result = self.handle_incoming(raw_message, handler=handler)
            response_message = result.get("response_message")
            if isinstance(response_message, dict):
                try:
                    self._publish_mqtt_response_message(
                        raw_message=raw_message,
                        response_message=response_message,
                    )
                except Exception:  # noqa: BLE001
                    return False
            return result.get("state") in terminal_states

        return self.mqtt_transport.consume(
            agent_id=self.agent_id,
            handler=_handle,
            mqtt_service=mqtt_service,
            max_messages=max_messages,
        )

    def request_capabilities(self, recipient: str) -> tuple[SendResult, dict[str, Any] | None]:
        result = self.send(
            recipients=[recipient],
            payload={"request": "capabilities"},
            message_class=MessageClass.CAPABILITIES,
            context=f"capabilities:{uuid.uuid4()}",
        )

        response_payload: dict[str, Any] | None = None
        for outcome in result.outcomes:
            if outcome.response_message is None:
                continue
            try:
                response_message, payload = self.decrypt_message_for_self(outcome.response_message)
            except Exception:  # noqa: BLE001
                continue
            if response_message.envelope.message_class is MessageClass.CAPABILITIES:
                response_payload = payload
                break
        return result, response_payload



# Copyright 2026 ACP Project
# Licensed under the Apache License, Version 2.0
# See LICENSE file for details.

from __future__ import annotations

import base64
import hashlib
import json
import os
from typing import Any, Mapping

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ed25519, x25519
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from cryptography.hazmat.primitives.kdf.hkdf import HKDF

from .messages import Envelope, ProtectedPayload, WrappedContentKey


class CryptoError(RuntimeError):
    pass


def b64encode(value: bytes) -> str:
    return base64.urlsafe_b64encode(value).decode("ascii")


def b64decode(value: str) -> bytes:
    return base64.urlsafe_b64decode(value.encode("ascii"))


def canonical_json(value: Any) -> str:
    return json.dumps(value, sort_keys=True, separators=(",", ":"), ensure_ascii=True)


def sha256_hex(value: bytes) -> str:
    return hashlib.sha256(value).hexdigest()


def generate_ed25519_keypair() -> tuple[str, str]:
    private_key = ed25519.Ed25519PrivateKey.generate()
    private_bytes = private_key.private_bytes(
        serialization.Encoding.Raw,
        serialization.PrivateFormat.Raw,
        serialization.NoEncryption(),
    )
    public_bytes = private_key.public_key().public_bytes(
        serialization.Encoding.Raw,
        serialization.PublicFormat.Raw,
    )
    return b64encode(private_bytes), b64encode(public_bytes)


def generate_x25519_keypair() -> tuple[str, str]:
    private_key = x25519.X25519PrivateKey.generate()
    private_bytes = private_key.private_bytes(
        serialization.Encoding.Raw,
        serialization.PrivateFormat.Raw,
        serialization.NoEncryption(),
    )
    public_bytes = private_key.public_key().public_bytes(
        serialization.Encoding.Raw,
        serialization.PublicFormat.Raw,
    )
    return b64encode(private_bytes), b64encode(public_bytes)


def _load_ed25519_private(private_key_b64: str) -> ed25519.Ed25519PrivateKey:
    return ed25519.Ed25519PrivateKey.from_private_bytes(b64decode(private_key_b64))


def _load_ed25519_public(public_key_b64: str) -> ed25519.Ed25519PublicKey:
    return ed25519.Ed25519PublicKey.from_public_bytes(b64decode(public_key_b64))


def _load_x25519_private(private_key_b64: str) -> x25519.X25519PrivateKey:
    return x25519.X25519PrivateKey.from_private_bytes(b64decode(private_key_b64))


def _load_x25519_public(public_key_b64: str) -> x25519.X25519PublicKey:
    return x25519.X25519PublicKey.from_public_bytes(b64decode(public_key_b64))


def sign_bytes(data: bytes, signing_private_key_b64: str) -> str:
    private_key = _load_ed25519_private(signing_private_key_b64)
    return b64encode(private_key.sign(data))


def verify_signature(data: bytes, signature_b64: str, signing_public_key_b64: str) -> bool:
    public_key = _load_ed25519_public(signing_public_key_b64)
    try:
        public_key.verify(b64decode(signature_b64), data)
        return True
    except InvalidSignature:
        return False


def envelope_aad(envelope: Envelope | dict[str, Any]) -> bytes:
    raw = envelope if isinstance(envelope, dict) else envelope.to_dict()
    aad = {
        "acp_version": raw["acp_version"],
        "message_id": raw["message_id"],
        "operation_id": raw["operation_id"],
        "sender": raw["sender"],
        "recipients": raw["recipients"],
    }
    return canonical_json(aad).encode("utf-8")


def _derive_wrap_key(shared_secret: bytes, recipient: str) -> bytes:
    hkdf = HKDF(
        algorithm=hashes.SHA256(),
        length=32,
        salt=None,
        info=f"acp-v1-wrap:{recipient}".encode("utf-8"),
    )
    return hkdf.derive(shared_secret)


def encrypt_for_recipients(
    payload: dict[str, Any],
    envelope: Envelope,
    recipient_encryption_public_keys: Mapping[str, str],
) -> ProtectedPayload:
    plaintext = canonical_json(payload).encode("utf-8")
    content_key = os.urandom(32)
    nonce = os.urandom(12)

    payload_aad = envelope_aad(envelope)
    ciphertext = AESGCM(content_key).encrypt(nonce, plaintext, payload_aad)

    ephemeral_private = x25519.X25519PrivateKey.generate()
    ephemeral_public_bytes = ephemeral_private.public_key().public_bytes(
        serialization.Encoding.Raw,
        serialization.PublicFormat.Raw,
    )

    wrapped_content_keys: list[WrappedContentKey] = []
    for recipient, recipient_public_key_b64 in recipient_encryption_public_keys.items():
        recipient_public = _load_x25519_public(recipient_public_key_b64)
        shared_secret = ephemeral_private.exchange(recipient_public)
        wrap_key = _derive_wrap_key(shared_secret, recipient)
        wrap_nonce = os.urandom(12)
        wrapped_cek = AESGCM(wrap_key).encrypt(
            wrap_nonce,
            content_key,
            envelope.message_id.encode("utf-8"),
        )
        wrapped_content_keys.append(
            WrappedContentKey(
                recipient=recipient,
                ephemeral_public_key=b64encode(ephemeral_public_bytes),
                nonce=b64encode(wrap_nonce),
                ciphertext=b64encode(wrapped_cek),
            )
        )

    return ProtectedPayload(
        nonce=b64encode(nonce),
        ciphertext=b64encode(ciphertext),
        wrapped_content_keys=wrapped_content_keys,
        payload_hash=sha256_hex(ciphertext),
        signature_kid="",
        signature="",
    )


def _message_signature_input(envelope: Envelope, protected: ProtectedPayload) -> bytes:
    body = {
        "envelope": envelope.to_dict(),
        "protected": protected.to_signable_dict(),
    }
    return canonical_json(body).encode("utf-8")


def sign_protected_payload(
    envelope: Envelope,
    protected: ProtectedPayload,
    signing_private_key_b64: str,
    signature_kid: str,
) -> ProtectedPayload:
    protected.signature_kid = signature_kid
    protected.signature = sign_bytes(
        _message_signature_input(envelope, protected),
        signing_private_key_b64,
    )
    return protected


def verify_protected_payload_signature(
    envelope: Envelope,
    protected: ProtectedPayload,
    sender_signing_public_key_b64: str,
) -> bool:
    if not protected.signature:
        return False
    return verify_signature(
        _message_signature_input(envelope, protected),
        protected.signature,
        sender_signing_public_key_b64,
    )


def decrypt_for_recipient(
    envelope: Envelope,
    protected: ProtectedPayload,
    recipient_id: str,
    recipient_encryption_private_key_b64: str,
) -> dict[str, Any]:
    matching = next(
        (item for item in protected.wrapped_content_keys if item.recipient == recipient_id),
        None,
    )
    if matching is None:
        raise CryptoError(f"No wrapped content key available for recipient {recipient_id}")

    recipient_private = _load_x25519_private(recipient_encryption_private_key_b64)
    ephemeral_public = _load_x25519_public(matching.ephemeral_public_key)
    shared_secret = recipient_private.exchange(ephemeral_public)
    wrap_key = _derive_wrap_key(shared_secret, recipient_id)

    try:
        content_key = AESGCM(wrap_key).decrypt(
            b64decode(matching.nonce),
            b64decode(matching.ciphertext),
            envelope.message_id.encode("utf-8"),
        )
    except Exception as exc:  # noqa: BLE001
        raise CryptoError("Failed to unwrap content key") from exc

    try:
        plaintext = AESGCM(content_key).decrypt(
            b64decode(protected.nonce),
            b64decode(protected.ciphertext),
            envelope_aad(envelope),
        )
    except Exception as exc:  # noqa: BLE001
        raise CryptoError("Failed to decrypt message payload") from exc

    return json.loads(plaintext.decode("utf-8"))



from acp.identity import AgentIdentity, verify_identity_document
import pytest


def test_identity_document_signature_verification() -> None:
    identity = AgentIdentity.create("agent:inventory.bot@localhost:9100")
    document = identity.build_identity_document(
        direct_endpoint="http://localhost:9100/acp/inbox",
        relay_hints=["http://localhost:8080"],
        trust_profile="domain_verified",
        capabilities={"agent_id": identity.agent_id},
    )
    assert verify_identity_document(document)

    tampered = dict(document)
    tampered["trust_profile"] = "self_asserted"
    assert not verify_identity_document(tampered)


def test_identity_document_rejects_unknown_trust_profile() -> None:
    identity = AgentIdentity.create("agent:inventory.bot@localhost:9100")
    with pytest.raises(ValueError):
        identity.build_identity_document(
            direct_endpoint="http://localhost:9100/acp/inbox",
            relay_hints=["http://localhost:8080"],
            trust_profile="unknown_profile",
            capabilities={"agent_id": identity.agent_id},
        )



# Copyright 2026 ACP Project
# Licensed under the Apache License, Version 2.0
# See LICENSE file for details.

from __future__ import annotations

import warnings
from typing import Any
import time

import requests

from .http_security import (
    HttpSecurityError,
    HttpSecurityPolicy,
    enforce_http_security,
    requests_cert_value,
    requests_verify_value,
    validate_http_security_policy,
)


class TransportError(RuntimeError):
    pass


class HTTPTransport:
    def __init__(
        self,
        *,
        timeout_seconds: int = 10,
        max_retries: int = 2,
        retry_backoff_seconds: float = 0.2,
        retry_status_codes: tuple[int, ...] | None = None,
        allow_insecure_http: bool = False,
        allow_insecure_tls: bool = False,
        ca_file: str | None = None,
        mtls_enabled: bool = False,
        cert_file: str | None = None,
        key_file: str | None = None,
    ) -> None:
        self.timeout_seconds = timeout_seconds
        self.max_retries = max_retries
        self.retry_backoff_seconds = retry_backoff_seconds
        self.retry_status_codes = retry_status_codes or (408, 425, 429, 500, 502, 503, 504)
        self.policy = HttpSecurityPolicy(
            allow_insecure_http=allow_insecure_http,
            allow_insecure_tls=allow_insecure_tls,
            ca_file=ca_file,
            mtls_enabled=mtls_enabled,
            cert_file=cert_file,
            key_file=key_file,
        )
        self._warned_messages: set[str] = set()
        try:
            warning_messages = validate_http_security_policy(
                self.policy,
                context="HTTP transport configuration",
            )
        except HttpSecurityError as exc:
            raise TransportError(str(exc)) from exc
        for warning_message in warning_messages:
            self._emit_warning(warning_message)

    def _emit_warning(self, message: str) -> None:
        if message in self._warned_messages:
            return
        self._warned_messages.add(message)
        warnings.warn(message, RuntimeWarning, stacklevel=3)

    def _validate_url(self, url: str, *, context: str) -> tuple[bool | str, tuple[str, str] | None]:
        try:
            warning_messages = enforce_http_security(url, policy=self.policy, context=context)
        except HttpSecurityError as exc:
            raise TransportError(str(exc)) from exc
        for warning_message in warning_messages:
            self._emit_warning(warning_message)
        try:
            cert = requests_cert_value(url, policy=self.policy)
        except HttpSecurityError as exc:
            raise TransportError(str(exc)) from exc
        return requests_verify_value(url, policy=self.policy), cert

    def _request_with_retries(
        self,
        method: str,
        url: str,
        *,
        json_body: dict[str, Any] | None = None,
        params: dict[str, str] | None = None,
    ) -> requests.Response:
        verify, cert = self._validate_url(url, context=f"HTTP {method.upper()} request")
        attempt = 0
        while True:
            try:
                response = requests.request(
                    method,
                    url,
                    json=json_body,
                    params=params,
                    timeout=self.timeout_seconds,
                    verify=verify,
                    cert=cert,
                )
            except requests.RequestException as exc:
                if attempt >= self.max_retries:
                    raise TransportError(f"HTTP {method} failed for {url}: {exc}") from exc
                time.sleep(self.retry_backoff_seconds * (2**attempt))
                attempt += 1
                continue

            if response.status_code in self.retry_status_codes and attempt < self.max_retries:
                time.sleep(self.retry_backoff_seconds * (2**attempt))
                attempt += 1
                continue
            return response

    def post_json(self, url: str, body: dict[str, Any]) -> requests.Response:
        return self._request_with_retries("POST", url, json_body=body)

    def get_json(self, url: str, params: dict[str, str] | None = None) -> dict[str, Any]:
        response = self._request_with_retries("GET", url, params=params)
        if response.status_code != 200:
            raise TransportError(f"HTTP GET {url} returned {response.status_code}")
        try:
            return response.json()
        except ValueError as exc:
            raise TransportError(f"HTTP GET {url} did not return JSON") from exc



from __future__ import annotations

from pathlib import Path
from typing import Any

import pytest

from acp.agent import Agent
from acp.identity import AgentIdentity, read_identity, write_identity
from acp.key_provider import IdentityKeyMaterial, KeyProviderError, LocalKeyProvider, TlsMaterial, VaultKeyProvider
from acp.transport import TransportError


class _FakeResponse:
    def __init__(self, status_code: int, payload: dict[str, Any]) -> None:
        self.status_code = status_code
        self._payload = payload

    def json(self) -> dict[str, Any]:
        return self._payload


def _seed_identity(storage_dir: Path, agent_id: str) -> AgentIdentity:
    identity = AgentIdentity.create(agent_id)
    document = identity.build_identity_document(
        direct_endpoint=None,
        relay_hints=[],
        trust_profile="self_asserted",
        capabilities={"agent_id": agent_id},
    )
    write_identity(storage_dir, identity, document)
    return identity


def test_local_key_provider_parity_with_existing_identity(tmp_path: Path) -> None:
    agent_id = "agent:key.local@demo"
    identity = _seed_identity(tmp_path, agent_id)
    provider = LocalKeyProvider(
        storage_dir=tmp_path,
        cert_file="/tmp/client-cert.pem",
        key_file="/tmp/client-key.pem",
        ca_file="/tmp/ca.pem",
    )

    keys = provider.load_identity_keys(agent_id)
    tls = provider.load_tls_material(agent_id)

    assert keys.signing_private_key == identity.signing_private_key
    assert keys.encryption_private_key == identity.encryption_private_key
    assert keys.signing_public_key == identity.signing_public_key
    assert keys.encryption_public_key == identity.encryption_public_key
    assert keys.signing_kid == identity.signing_kid
    assert keys.encryption_kid == identity.encryption_kid
    assert tls == TlsMaterial(
        cert_file="/tmp/client-cert.pem",
        key_file="/tmp/client-key.pem",
        ca_file="/tmp/ca.pem",
    )
    assert provider.load_ca_bundle(agent_id) == "/tmp/ca.pem"
    assert provider.describe()["provider"] == "local"


def test_local_key_provider_missing_identity_raises(tmp_path: Path) -> None:
    provider = LocalKeyProvider(storage_dir=tmp_path)
    with pytest.raises(KeyProviderError, match="Local identity not found"):
        provider.load_identity_keys("agent:missing@demo")


def test_vault_key_provider_happy_path(monkeypatch: pytest.MonkeyPatch) -> None:
    captured: dict[str, Any] = {}

    def fake_get(url: str, *, headers: dict[str, str], timeout: int, verify: bool | str) -> _FakeResponse:
        captured["url"] = url
        captured["headers"] = headers
        captured["timeout"] = timeout
        captured["verify"] = verify
        return _FakeResponse(
            200,
            {
                "data": {
                    "data": {
                        "signing_key": "sig-private",
                        "encryption_key": "enc-private",
                        "signing_public_key": "sig-public",
                        "encryption_public_key": "enc-public",
                        "signing_kid": "sig-kid",
                        "encryption_kid": "enc-kid",
                        "tls_cert_file": "/etc/acp/client-cert.pem",
                        "tls_key_file": "/etc/acp/client-key.pem",
                        "ca_file": "/etc/acp/ca.pem",
                    },
                },
            },
        )

    monkeypatch.setenv("VAULT_TOKEN", "token-123")
    monkeypatch.setattr("acp.key_provider.requests.get", fake_get)

    provider = VaultKeyProvider(
        vault_url="https://vault.local",
        vault_path="secret/data/acp/identities",
        vault_token_env="VAULT_TOKEN",
        timeout_seconds=7,
    )
    keys = provider.load_identity_keys("agent:john.chess@demo")
    tls = provider.load_tls_material("agent:john.chess@demo")

    assert captured["url"] == "https://vault.local/v1/secret/data/acp/identities/agent_john.chess_demo"
    assert captured["headers"]["X-Vault-Token"] == "token-123"
    assert captured["timeout"] == 7
    assert captured["verify"] is True
    assert keys == IdentityKeyMaterial(
        signing_private_key="sig-private",
        encryption_private_key="enc-private",
        signing_public_key="sig-public",
        encryption_public_key="enc-public",
        signing_kid="sig-kid",
        encryption_kid="enc-kid",
    )
    assert tls == TlsMaterial(
        cert_file="/etc/acp/client-cert.pem",
        key_file="/etc/acp/client-key.pem",
        ca_file="/etc/acp/ca.pem",
    )
    assert provider.load_ca_bundle("agent:john.chess@demo") == "/etc/acp/ca.pem"


def test_vault_key_provider_missing_token_raises(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.delenv("VAULT_TOKEN", raising=False)
    provider = VaultKeyProvider(
        vault_url="https://vault.local",
        vault_path="secret/data/acp/identities",
    )
    with pytest.raises(KeyProviderError, match="Vault token is missing"):
        provider.load_identity_keys("agent:john@demo")


def test_vault_key_provider_missing_required_fields_raises(monkeypatch: pytest.MonkeyPatch) -> None:
    def fake_get(_url: str, *, headers: dict[str, str], timeout: int, verify: bool | str) -> _FakeResponse:
        _ = headers, timeout, verify
        return _FakeResponse(200, {"data": {"data": {"encryption_key": "enc-only"}}})

    monkeypatch.setenv("VAULT_TOKEN", "token-123")
    monkeypatch.setattr("acp.key_provider.requests.get", fake_get)
    provider = VaultKeyProvider(
        vault_url="https://vault.local",
        vault_path="secret/data/acp/identities",
    )
    with pytest.raises(KeyProviderError, match="missing signing_key"):
        provider.load_identity_keys("agent:john@demo")


def test_agent_load_or_create_uses_custom_key_provider(tmp_path: Path) -> None:
    bootstrap_identity = AgentIdentity.create("agent:provider.bootstrap@demo")

    class FakeProvider:
        def load_identity_keys(self, agent_id: str) -> IdentityKeyMaterial:
            assert agent_id == "agent:provider.bootstrap@demo"
            return IdentityKeyMaterial(
                signing_private_key=bootstrap_identity.signing_private_key,
                encryption_private_key=bootstrap_identity.encryption_private_key,
                signing_public_key=bootstrap_identity.signing_public_key,
                encryption_public_key=bootstrap_identity.encryption_public_key,
                signing_kid=bootstrap_identity.signing_kid,
                encryption_kid=bootstrap_identity.encryption_kid,
            )

        def load_tls_material(self, agent_id: str | None = None) -> TlsMaterial:
            _ = agent_id
            return TlsMaterial()

        def load_ca_bundle(self, agent_id: str | None = None) -> str | None:
            _ = agent_id
            return None

        def describe(self) -> dict[str, Any]:
            return {"provider": "fake", "path": "secret/acp/identities"}

    agent = Agent.load_or_create(
        "agent:provider.bootstrap@demo",
        storage_dir=tmp_path / "agent",
        key_provider=FakeProvider(),
        relay_url="https://relay.example",
    )
    assert agent.key_provider_info["provider"] == "fake"
    assert agent.identity.signing_private_key == bootstrap_identity.signing_private_key
    assert agent.identity.encryption_private_key == bootstrap_identity.encryption_private_key
    assert read_identity(tmp_path / "agent", "agent:provider.bootstrap@demo") is not None


def test_agent_load_or_create_external_provider_requires_public_metadata(tmp_path: Path) -> None:
    local_identity = AgentIdentity.create("agent:provider.incomplete@demo")

    class IncompleteProvider:
        def load_identity_keys(self, _agent_id: str) -> IdentityKeyMaterial:
            return IdentityKeyMaterial(
                signing_private_key=local_identity.signing_private_key,
                encryption_private_key=local_identity.encryption_private_key,
            )

        def load_tls_material(self, agent_id: str | None = None) -> TlsMaterial:
            _ = agent_id
            return TlsMaterial()

        def load_ca_bundle(self, agent_id: str | None = None) -> str | None:
            _ = agent_id
            return None

        def describe(self) -> dict[str, Any]:
            return {"provider": "incomplete"}

    with pytest.raises(TransportError, match="first-time bootstrap"):
        Agent.load_or_create(
            "agent:provider.incomplete@demo",
            storage_dir=tmp_path / "agent",
            key_provider=IncompleteProvider(),
            relay_url="https://relay.example",
        )



# Copyright 2026 ACP Project
# Licensed under the Apache License, Version 2.0
# See LICENSE file for details.

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime, timedelta, timezone
from enum import Enum
import json
import uuid
from typing import Any


ACP_VERSION = "1.0"
DEFAULT_CRYPTO_SUITE = "ACP-AES256-GCM+X25519+ED25519"


def utc_now_iso() -> str:
    return datetime.now(timezone.utc).isoformat().replace("+00:00", "Z")


def parse_iso8601(value: str) -> datetime:
    return datetime.fromisoformat(value.replace("Z", "+00:00"))


class MessageClass(str, Enum):
    SEND = "SEND"
    ACK = "ACK"
    FAIL = "FAIL"
    CAPABILITIES = "CAPABILITIES"
    COMPENSATE = "COMPENSATE"


class DeliveryState(str, Enum):
    PENDING = "PENDING"
    DELIVERED = "DELIVERED"
    ACKNOWLEDGED = "ACKNOWLEDGED"
    FAILED = "FAILED"
    DECLINED = "DECLINED"
    EXPIRED = "EXPIRED"


class FailReason(str, Enum):
    UNSUPPORTED_VERSION = "UNSUPPORTED_VERSION"
    UNSUPPORTED_CRYPTO_SUITE = "UNSUPPORTED_CRYPTO_SUITE"
    UNSUPPORTED_MESSAGE_CLASS = "UNSUPPORTED_MESSAGE_CLASS"
    INVALID_SIGNATURE = "INVALID_SIGNATURE"
    EXPIRED_MESSAGE = "EXPIRED_MESSAGE"
    POLICY_REJECTED = "POLICY_REJECTED"
    PAYLOAD_TOO_LARGE = "PAYLOAD_TOO_LARGE"
    UNSUPPORTED_PROFILE = "UNSUPPORTED_PROFILE"


@dataclass
class WrappedContentKey:
    recipient: str
    ephemeral_public_key: str
    nonce: str
    ciphertext: str

    def to_dict(self) -> dict[str, str]:
        return {
            "recipient": self.recipient,
            "ephemeral_public_key": self.ephemeral_public_key,
            "nonce": self.nonce,
            "ciphertext": self.ciphertext,
        }

    @classmethod
    def from_dict(cls, value: dict[str, Any]) -> "WrappedContentKey":
        return cls(
            recipient=str(value["recipient"]),
            ephemeral_public_key=str(value["ephemeral_public_key"]),
            nonce=str(value["nonce"]),
            ciphertext=str(value["ciphertext"]),
        )


@dataclass
class Envelope:
    acp_version: str
    message_class: MessageClass
    message_id: str
    operation_id: str
    timestamp: str
    expires_at: str
    sender: str
    recipients: list[str]
    context_id: str
    crypto_suite: str = DEFAULT_CRYPTO_SUITE
    correlation_id: str | None = None
    in_reply_to: str | None = None

    @classmethod
    def build(
        cls,
        *,
        sender: str,
        recipients: list[str],
        message_class: MessageClass,
        context_id: str,
        expires_in_seconds: int = 300,
        operation_id: str | None = None,
        correlation_id: str | None = None,
        in_reply_to: str | None = None,
        crypto_suite: str = DEFAULT_CRYPTO_SUITE,
    ) -> "Envelope":
        now = datetime.now(timezone.utc)
        operation = operation_id or str(uuid.uuid4())
        return cls(
            acp_version=ACP_VERSION,
            message_class=message_class,
            message_id=str(uuid.uuid4()),
            operation_id=operation,
            timestamp=now.isoformat().replace("+00:00", "Z"),
            expires_at=(now + timedelta(seconds=expires_in_seconds))
            .isoformat()
            .replace("+00:00", "Z"),
            sender=sender,
            recipients=recipients,
            context_id=context_id,
            crypto_suite=crypto_suite,
            correlation_id=correlation_id,
            in_reply_to=in_reply_to,
        )

    def validate(self) -> None:
        if not self.recipients:
            raise ValueError("Envelope recipients must not be empty")
        if not self.sender:
            raise ValueError("Envelope sender is required")
        if parse_iso8601(self.expires_at) <= parse_iso8601(self.timestamp):
            raise ValueError("Envelope expires_at must be after timestamp")

    def is_expired(self) -> bool:
        return parse_iso8601(self.expires_at) <= datetime.now(timezone.utc)

    def to_dict(self) -> dict[str, Any]:
        self.validate()
        value: dict[str, Any] = {
            "acp_version": self.acp_version,
            "message_class": self.message_class.value,
            "message_id": self.message_id,
            "operation_id": self.operation_id,
            "timestamp": self.timestamp,
            "expires_at": self.expires_at,
            "sender": self.sender,
            "recipients": self.recipients,
            "context_id": self.context_id,
            "crypto_suite": self.crypto_suite,
        }
        if self.correlation_id is not None:
            value["correlation_id"] = self.correlation_id
        if self.in_reply_to is not None:
            value["in_reply_to"] = self.in_reply_to
        return value

    @classmethod
    def from_dict(cls, value: dict[str, Any]) -> "Envelope":
        envelope = cls(
            acp_version=str(value["acp_version"]),
            message_class=MessageClass(value["message_class"]),
            message_id=str(value["message_id"]),
            operation_id=str(value["operation_id"]),
            timestamp=str(value["timestamp"]),
            expires_at=str(value["expires_at"]),
            sender=str(value["sender"]),
            recipients=[str(item) for item in value["recipients"]],
            context_id=str(value["context_id"]),
            crypto_suite=str(value["crypto_suite"]),
            correlation_id=(
                str(value["correlation_id"])
                if value.get("correlation_id") is not None
                else None
            ),
            in_reply_to=(
                str(value["in_reply_to"])
                if value.get("in_reply_to") is not None
                else None
            ),
        )
        envelope.validate()
        return envelope


@dataclass
class ProtectedPayload:
    nonce: str
    ciphertext: str
    wrapped_content_keys: list[WrappedContentKey]
    payload_hash: str
    signature_kid: str
    signature: str

    def to_dict(self) -> dict[str, Any]:
        return {
            "nonce": self.nonce,
            "ciphertext": self.ciphertext,
            "wrapped_content_keys": [item.to_dict() for item in self.wrapped_content_keys],
            "payload_hash": self.payload_hash,
            "signature_kid": self.signature_kid,
            "signature": self.signature,
        }

    def to_signable_dict(self) -> dict[str, Any]:
        keys = sorted(
            (item.to_dict() for item in self.wrapped_content_keys),
            key=lambda item: item["recipient"],
        )
        return {
            "nonce": self.nonce,
            "ciphertext": self.ciphertext,
            "wrapped_content_keys": keys,
            "payload_hash": self.payload_hash,
            "signature_kid": self.signature_kid,
        }

    @classmethod
    def from_dict(cls, value: dict[str, Any]) -> "ProtectedPayload":
        return cls(
            nonce=str(value["nonce"]),
            ciphertext=str(value["ciphertext"]),
            wrapped_content_keys=[
                WrappedContentKey.from_dict(item) for item in value["wrapped_content_keys"]
            ],
            payload_hash=str(value["payload_hash"]),
            signature_kid=str(value["signature_kid"]),
            signature=str(value["signature"]),
        )


@dataclass
class ACPMessage:
    envelope: Envelope
    protected: ProtectedPayload
    sender_identity_document: dict[str, Any] | None = None

    def to_dict(self) -> dict[str, Any]:
        value: dict[str, Any] = {
            "envelope": self.envelope.to_dict(),
            "protected": self.protected.to_dict(),
        }
        if self.sender_identity_document is not None:
            value["sender_identity_document"] = self.sender_identity_document
        return value

    def to_json(self) -> str:
        return json.dumps(self.to_dict(), sort_keys=True, separators=(",", ":"))

    @classmethod
    def from_dict(cls, value: dict[str, Any]) -> "ACPMessage":
        return cls(
            envelope=Envelope.from_dict(value["envelope"]),
            protected=ProtectedPayload.from_dict(value["protected"]),
            sender_identity_document=value.get("sender_identity_document"),
        )

    @classmethod
    def from_json(cls, value: str) -> "ACPMessage":
        return cls.from_dict(json.loads(value))


@dataclass
class DeliveryOutcome:
    recipient: str
    state: DeliveryState
    status_code: int | None = None
    response_class: MessageClass | None = None
    reason_code: str | None = None
    detail: str | None = None
    response_message: dict[str, Any] | None = None

    def to_dict(self) -> dict[str, Any]:
        value: dict[str, Any] = {
            "recipient": self.recipient,
            "state": self.state.value,
        }
        if self.status_code is not None:
            value["status_code"] = self.status_code
        if self.response_class is not None:
            value["response_class"] = self.response_class.value
        if self.reason_code is not None:
            value["reason_code"] = self.reason_code
        if self.detail is not None:
            value["detail"] = self.detail
        if self.response_message is not None:
            value["response_message"] = self.response_message
        return value

    @classmethod
    def from_dict(cls, value: dict[str, Any]) -> "DeliveryOutcome":
        response_class = value.get("response_class")
        return cls(
            recipient=str(value["recipient"]),
            state=DeliveryState(str(value["state"])),
            status_code=value.get("status_code"),
            response_class=MessageClass(response_class) if response_class else None,
            reason_code=value.get("reason_code"),
            detail=value.get("detail"),
            response_message=value.get("response_message"),
        )


@dataclass
class SendResult:
    operation_id: str
    message_id: str
    outcomes: list[DeliveryOutcome]
    message_ids: list[str] | None = None

    def to_dict(self) -> dict[str, Any]:
        value: dict[str, Any] = {
            "operation_id": self.operation_id,
            "message_id": self.message_id,
            "outcomes": [outcome.to_dict() for outcome in self.outcomes],
        }
        if self.message_ids:
            value["message_ids"] = self.message_ids
        return value


@dataclass
class CompensateInstruction:
    operation_id: str
    reason: str
    actions: list[dict[str, Any]] = field(default_factory=list)

    def to_dict(self) -> dict[str, Any]:
        return {
            "operation_id": self.operation_id,
            "reason": self.reason,
            "actions": self.actions,
        }

    @classmethod
    def from_dict(cls, value: dict[str, Any]) -> "CompensateInstruction":
        return cls(
            operation_id=str(value["operation_id"]),
            reason=str(value["reason"]),
            actions=[dict(item) for item in value.get("actions", [])],
        )


def build_ack_payload(received_message_id: str, status: str = "accepted") -> dict[str, Any]:
    return {
        "status": status,
        "received_message_id": received_message_id,
    }


def build_fail_payload(
    reason_code: str,
    detail: str,
    retriable: bool = False,
) -> dict[str, Any]:
    return {
        "reason_code": reason_code,
        "detail": detail,
        "retriable": retriable,
    }



from __future__ import annotations

from copy import deepcopy
from pathlib import Path
from typing import Any

from acp import Agent, DeliveryState
from acp.amqp_transport import (
    build_amqp_service_hint,
    queue_name_for_agent,
    routing_key_for_agent,
)
from acp.messages import MessageClass


class FakeAMQPTransport:
    def __init__(self) -> None:
        self.published: list[dict[str, Any]] = []

    def publish(
        self,
        *,
        message: dict[str, Any],
        recipient_agent_id: str,
        amqp_service: dict[str, Any] | None = None,
    ) -> None:
        self.published.append(
            {
                "recipient": recipient_agent_id,
                "message": message,
                "amqp_service": dict(amqp_service or {}),
            },
        )


def _recipient_doc_with_amqp(recipient: Agent, broker_url: str) -> dict[str, Any]:
    doc = deepcopy(recipient.identity_document)
    service = dict(doc.get("service", {}))
    service["direct_endpoint"] = None
    service["amqp"] = build_amqp_service_hint(
        agent_id=recipient.agent_id,
        broker_url=broker_url,
        exchange="acp.exchange",
    )
    doc["service"] = service

    capabilities = dict(doc.get("capabilities", {}))
    capabilities["transports"] = ["amqp"]
    doc["capabilities"] = capabilities
    return doc


def _seed_local_amqp_service(agent: Agent, broker_url: str) -> None:
    service = dict(agent.identity_document.get("service", {}))
    service["amqp"] = build_amqp_service_hint(
        agent_id=agent.agent_id,
        broker_url=broker_url,
        exchange="acp.exchange",
    )
    agent.identity_document["service"] = service


def test_amqp_queue_and_routing_key_conventions() -> None:
    agent_id = "agent:shipping.bot@companyB.com"
    assert queue_name_for_agent(agent_id) == "acp.agent.shipping.bot.companyB.com"
    assert routing_key_for_agent(agent_id) == "agent.shipping.bot.companyB.com"


def test_send_amqp_mode_publishes_one_message_per_recipient(tmp_path: Path) -> None:
    sender = Agent.create(
        "agent:sender.bot@localhost:9400",
        storage_dir=tmp_path / "sender",
        endpoint="http://localhost:9400/acp/inbox",
        relay_url="http://localhost:8080",
        relay_hints=["http://localhost:8080"],
        discovery_scheme="http",
        allow_insecure_http=True,
    )
    recipient1 = Agent.create(
        "agent:recipient1.bot@localhost:9401",
        storage_dir=tmp_path / "recipient1",
        endpoint="http://localhost:9401/acp/inbox",
        relay_url="http://localhost:8080",
        relay_hints=["http://localhost:8080"],
        discovery_scheme="http",
        allow_insecure_http=True,
    )
    recipient2 = Agent.create(
        "agent:recipient2.bot@localhost:9402",
        storage_dir=tmp_path / "recipient2",
        endpoint="http://localhost:9402/acp/inbox",
        relay_url="http://localhost:8080",
        relay_hints=["http://localhost:8080"],
        discovery_scheme="http",
        allow_insecure_http=True,
    )

    fake_transport = FakeAMQPTransport()
    sender.amqp_transport = fake_transport  # type: ignore[assignment]
    recipient_docs = {
        recipient1.agent_id: _recipient_doc_with_amqp(recipient1, "amqp://broker.local"),
        recipient2.agent_id: _recipient_doc_with_amqp(recipient2, "amqp://broker.local"),
    }

    sender.discovery.resolve = lambda agent_id: recipient_docs[agent_id]  # type: ignore[method-assign]

    result = sender.send(
        recipients=[recipient1.agent_id, recipient2.agent_id],
        payload={"type": "hand_start", "hand_id": "h-1"},
        context="ctx-amqp",
        delivery_mode="amqp",
    )

    assert len(fake_transport.published) == 2
    published_recipients = {item["recipient"] for item in fake_transport.published}
    assert published_recipients == {recipient1.agent_id, recipient2.agent_id}
    for item in fake_transport.published:
        envelope = item["message"]["envelope"]
        assert envelope["recipients"] == [item["recipient"]]
        assert item["amqp_service"]["routing_key"] == routing_key_for_agent(item["recipient"])

    assert len(result.message_ids or []) == 2
    assert all(outcome.state is DeliveryState.DELIVERED for outcome in result.outcomes)


def test_consume_from_amqp_acknowledges_duplicate_delivery(tmp_path: Path) -> None:
    sender = Agent.create(
        "agent:sender.bot@localhost:9500",
        storage_dir=tmp_path / "sender",
        endpoint="http://localhost:9500/acp/inbox",
        discovery_scheme="http",
        allow_insecure_http=True,
    )
    receiver = Agent.create(
        "agent:receiver.bot@localhost:9501",
        storage_dir=tmp_path / "receiver",
        endpoint="http://localhost:9501/acp/inbox",
        discovery_scheme="http",
        allow_insecure_http=True,
    )

    _seed_local_amqp_service(sender, "amqp://broker.local")
    _seed_local_amqp_service(receiver, "amqp://broker.local")
    receiver_public_key = receiver.identity_document["keys"]["encryption"]["public_key"]
    outbound = sender._build_message(  # noqa: SLF001
        recipients=[receiver.agent_id],
        payload={"type": "ping"},
        recipient_public_keys={receiver.agent_id: receiver_public_key},
        message_class=MessageClass.SEND,
        context_id="ctx-consume",
        operation_id="op-consume",
        expires_in_seconds=60,
        correlation_id=None,
        in_reply_to=None,
    ).to_dict()

    class DuplicateDeliveryTransport:
        def __init__(self) -> None:
            self.acks: list[bool] = []
            self.published: list[dict[str, Any]] = []

        def publish(
            self,
            *,
            message: dict[str, Any],
            recipient_agent_id: str,
            amqp_service: dict[str, Any] | None = None,
        ) -> None:
            self.published.append(
                {
                    "recipient": recipient_agent_id,
                    "message": message,
                    "amqp_service": dict(amqp_service or {}),
                },
            )

        def consume(
            self,
            *,
            agent_id: str,
            handler: Any,
            amqp_service: dict[str, Any] | None = None,
            max_messages: int | None = None,
        ) -> int:
            self.acks.append(bool(handler(outbound)))
            self.acks.append(bool(handler(outbound)))
            return 2

    fake_consumer = DuplicateDeliveryTransport()
    receiver.amqp_transport = fake_consumer  # type: ignore[assignment]
    receiver.discovery.resolve = lambda agent_id: sender.identity_document  # type: ignore[method-assign]

    consumed = receiver.consume_from_amqp(max_messages=2)
    assert consumed == 2
    assert fake_consumer.acks == [True, True]
    assert len(fake_consumer.published) == 2
    assert all(
        item["message"]["envelope"]["message_class"] == MessageClass.ACK.value
        for item in fake_consumer.published
    )


def test_consume_from_amqp_publishes_response_ack_to_sender(tmp_path: Path) -> None:
    sender = Agent.create(
        "agent:sender.bot@localhost:9600",
        storage_dir=tmp_path / "sender",
        endpoint="http://localhost:9600/acp/inbox",
        discovery_scheme="http",
        allow_insecure_http=True,
    )
    receiver = Agent.create(
        "agent:receiver.bot@localhost:9601",
        storage_dir=tmp_path / "receiver",
        endpoint="http://localhost:9601/acp/inbox",
        discovery_scheme="http",
        allow_insecure_http=True,
    )
    _seed_local_amqp_service(sender, "amqp://broker.local")
    _seed_local_amqp_service(receiver, "amqp://broker.local")

    receiver_public_key = receiver.identity_document["keys"]["encryption"]["public_key"]
    outbound = sender._build_message(  # noqa: SLF001
        recipients=[receiver.agent_id],
        payload={"type": "ping"},
        recipient_public_keys={receiver.agent_id: receiver_public_key},
        message_class=MessageClass.SEND,
        context_id="ctx-consume-send",
        operation_id="op-consume-send",
        expires_in_seconds=60,
        correlation_id=None,
        in_reply_to=None,
    ).to_dict()

    class RoundTripTransport:
        def __init__(self) -> None:
            self.acks: list[bool] = []
            self.published: list[dict[str, Any]] = []

        def publish(
            self,
            *,
            message: dict[str, Any],
            recipient_agent_id: str,
            amqp_service: dict[str, Any] | None = None,
        ) -> None:
            self.published.append(
                {
                    "recipient": recipient_agent_id,
                    "message": message,
                    "amqp_service": dict(amqp_service or {}),
                },
            )

        def consume(
            self,
            *,
            agent_id: str,
            handler: Any,
            amqp_service: dict[str, Any] | None = None,
            max_messages: int | None = None,
        ) -> int:
            self.acks.append(bool(handler(outbound)))
            return 1

    fake_transport = RoundTripTransport()
    receiver.amqp_transport = fake_transport  # type: ignore[assignment]
    receiver.discovery.resolve = lambda agent_id: sender.identity_document  # type: ignore[method-assign]

    consumed = receiver.consume_from_amqp(max_messages=1)
    assert consumed == 1
    assert fake_transport.acks == [True]
    assert len(fake_transport.published) == 1
    published = fake_transport.published[0]
    assert published["recipient"] == sender.agent_id
    assert published["message"]["envelope"]["message_class"] == MessageClass.ACK.value
    assert published["amqp_service"]["routing_key"] == routing_key_for_agent(sender.agent_id)


def test_handle_incoming_ack_message_does_not_generate_ack_of_ack(tmp_path: Path) -> None:
    sender = Agent.create(
        "agent:sender.bot@localhost:9700",
        storage_dir=tmp_path / "sender",
        endpoint="http://localhost:9700/acp/inbox",
        discovery_scheme="http",
        allow_insecure_http=True,
    )
    receiver = Agent.create(
        "agent:receiver.bot@localhost:9701",
        storage_dir=tmp_path / "receiver",
        endpoint="http://localhost:9701/acp/inbox",
        discovery_scheme="http",
        allow_insecure_http=True,
    )

    sender_public_key = sender.identity_document["keys"]["encryption"]["public_key"]
    ack_message = receiver._build_message(  # noqa: SLF001
        recipients=[sender.agent_id],
        payload={"status": "accepted", "received_message_id": "m-original"},
        recipient_public_keys={sender.agent_id: sender_public_key},
        message_class=MessageClass.ACK,
        context_id="ctx-ack",
        operation_id="op-ack",
        expires_in_seconds=60,
        correlation_id="op-original",
        in_reply_to="m-original",
    ).to_dict()

    sender.discovery.resolve = lambda agent_id: receiver.identity_document  # type: ignore[method-assign]
    result = sender.handle_incoming(ack_message)
    assert result["state"] == DeliveryState.ACKNOWLEDGED.value
    assert result["response_message"] is None
